import unittest
import pandas as pd
from pandas.testing import assert_frame_equal
from datetime import datetime
from src.nemosis import defaults, query_wrappers


class TestDispatchDateSetup(unittest.TestCase):
    def setUp(self):
        self.start_time = "2017/01/01 00:10:00"
        self.end_time = "2019/06/03 12:15:00"

    def test_start_time_and_end_time(self):
        start_time, end_time = query_wrappers.dispatch_date_setup(
            self.start_time, self.end_time
        )
        self.assertEqual(start_time, "2016/12/30 23:59:59")
        self.assertEqual(end_time, "2019/06/03 00:00:00")


class TestFCASFinalise(unittest.TestCase):
    def setUp(self):
        self.testDataFrame1 = pd.DataFrame(
            {"StingsWithSpaces": [" HELLO ", "foo and baa ", " at front"]}
        )
        self.testDataFrame2 = pd.DataFrame(
            {
                "StingsWithSpaces": [" HELLO ", "foo and baa ", " at front"],
                "INTs2ignore": [1, 2, 3],
            }
        )
        self.testDataFrame3 = pd.DataFrame(
            {
                "StingsWithSpaces": [" HELLO ", "foo and baa ", " at front"],
                "INTs2ignore": [1, 2, 3],
                "StingsWithSpaces2": [" HELLO ", "foo and baa ", " at front"],
            }
        )

    def test_string_cleanup_one_col(self):
        result = query_wrappers.fcas4s_finalise(self.testDataFrame1, None, None)
        aim = pd.DataFrame({"StingsWithSpaces": ["HELLO", "foo and baa", "at front"]})
        assert_frame_equal(aim, result)

    def test_string_cleanup_one_col_ignore_one_col(self):
        result = query_wrappers.fcas4s_finalise(self.testDataFrame2, None, None)
        aim = pd.DataFrame(
            {
                "StingsWithSpaces": ["HELLO", "foo and baa", "at front"],
                "INTs2ignore": [1, 2, 3],
            }
        )
        assert_frame_equal(aim, result)

    def test_cleanup_ignore_cleanup(self):
        result = query_wrappers.fcas4s_finalise(self.testDataFrame3, None, None)
        aim = pd.DataFrame(
            {
                "StingsWithSpaces": ["HELLO", "foo and baa", "at front"],
                "INTs2ignore": [1, 2, 3],
                "StingsWithSpaces2": ["HELLO", "foo and baa", "at front"],
            }
        )
        assert_frame_equal(aim, result)


class TestMostRecent(unittest.TestCase):
    def setUp(self):
        self.dummyGenConData = pd.DataFrame(
            {
                "EFFECTIVEDATE": [
                    "2017/01/01 00:00:00",
                    "2017/01/04 00:15:00",
                    "2018/05/01 00:00:00",
                ],
                "VERSIONNO": ["5", "1", "1"],
                "GENCONID": ["ID1", "ID1", "ID1"],
            }
        )
        self.dummyGenConData["EFFECTIVEDATE"] = pd.to_datetime(
            self.dummyGenConData["EFFECTIVEDATE"], format="%Y/%m/%d %H:%M:%S"
        )

        self.dummyGenConData2 = pd.DataFrame(
            {
                "EFFECTIVEDATE": [
                    "2017/01/01 00:00:00",
                    "2017/01/04 00:15:00",
                    "2017/01/04 00:15:00",
                ],
                "VERSIONNO": ["5", "1", "2"],
                "GENCONID": ["ID1", "ID1", "ID1"],
            }
        )
        self.dummyGenConData2["EFFECTIVEDATE"] = pd.to_datetime(
            self.dummyGenConData2["EFFECTIVEDATE"], format="%Y/%m/%d %H:%M:%S"
        )

        self.dummyGenConData3 = pd.DataFrame(
            {
                "EFFECTIVEDATE": [
                    "2017/01/01 00:00:00",
                    "2017/01/04 00:15:00",
                    "2017/01/04 00:15:00",
                    "2017/01/01 00:00:00",
                    "2017/01/04 00:15:00",
                    "2017/01/04 00:15:00",
                ],
                "VERSIONNO": ["5", "1", "2", "5", "1", "2"],
                "GENCONID": ["ID1", "ID1", "ID1", "ID2", "ID2", "ID2"],
            }
        )
        self.dummyGenConData3["EFFECTIVEDATE"] = pd.to_datetime(
            self.dummyGenConData3["EFFECTIVEDATE"], format="%Y/%m/%d %H:%M:%S"
        )

    def test_one_id_gencondata_start_date_after_all(self):
        start_time = datetime.strptime("2019/06/01 00:00:00", "%Y/%m/%d %H:%M:%S")
        table_name = "dummy"
        defaults.primary_date_columns["dummy"] = "EFFECTIVEDATE"
        defaults.effective_date_group_col["dummy"] = ["GENCONID"]
        result = query_wrappers.most_recent_records_before_start_time(
            self.dummyGenConData, start_time, table_name
        ).reset_index(drop=True)
        aim = pd.DataFrame(
            {
                "EFFECTIVEDATE": ["2018/05/01 00:00:00"],
                "VERSIONNO": ["1"],
                "GENCONID": ["ID1"],
            }
        )
        aim["EFFECTIVEDATE"] = pd.to_datetime(
            aim["EFFECTIVEDATE"], format="%Y/%m/%d %H:%M:%S"
        )
        assert_frame_equal(aim, result)

    def test_one_id_gencondata_start_date_after_first(self):
        start_time = datetime.strptime("2017/01/01 01:00:00", "%Y/%m/%d %H:%M:%S")
        table_name = "dummy"
        defaults.primary_date_columns["dummy"] = "EFFECTIVEDATE"
        defaults.effective_date_group_col["dummy"] = ["GENCONID"]
        result = (
            query_wrappers.most_recent_records_before_start_time(
                self.dummyGenConData, start_time, table_name
            )
            .sort_values("EFFECTIVEDATE", ascending=False)
            .reset_index(drop=True)
        )
        aim = (
            pd.DataFrame(
                {
                    "EFFECTIVEDATE": [
                        "2017/01/01 00:00:00",
                        "2017/01/04 00:15:00",
                        "2018/05/01 00:00:00",
                    ],
                    "VERSIONNO": ["5", "1", "1"],
                    "GENCONID": ["ID1", "ID1", "ID1"],
                }
            )
            .sort_values("EFFECTIVEDATE", ascending=False)
            .reset_index(drop=True)
        )
        aim["EFFECTIVEDATE"] = pd.to_datetime(
            aim["EFFECTIVEDATE"], format="%Y/%m/%d %H:%M:%S"
        )
        assert_frame_equal(aim, result)

    def test_one_id_gencondata_start_date_before_all(self):
        start_time = datetime.strptime("2010/01/01 01:00:00", "%Y/%m/%d %H:%M:%S")
        table_name = "dummy"
        defaults.primary_date_columns["dummy"] = "EFFECTIVEDATE"
        defaults.effective_date_group_col["dummy"] = ["GENCONID"]
        result = (
            query_wrappers.most_recent_records_before_start_time(
                self.dummyGenConData, start_time, table_name
            )
            .sort_values("EFFECTIVEDATE", ascending=False)
            .reset_index(drop=True)
        )
        aim = (
            pd.DataFrame(
                {
                    "EFFECTIVEDATE": [
                        "2017/01/01 00:00:00",
                        "2017/01/04 00:15:00",
                        "2018/05/01 00:00:00",
                    ],
                    "VERSIONNO": ["5", "1", "1"],
                    "GENCONID": ["ID1", "ID1", "ID1"],
                }
            )
            .sort_values("EFFECTIVEDATE", ascending=False)
            .reset_index(drop=True)
        )
        aim["EFFECTIVEDATE"] = pd.to_datetime(
            aim["EFFECTIVEDATE"], format="%Y/%m/%d %H:%M:%S"
        )
        assert_frame_equal(aim, result)

    def test_one_id_gencondata_and_repeated_effectivedate_start_date_after_all(self):
        start_time = datetime.strptime("2019/06/01 00:00:00", "%Y/%m/%d %H:%M:%S")
        table_name = "dummy"
        defaults.primary_date_columns["dummy"] = "EFFECTIVEDATE"
        defaults.effective_date_group_col["dummy"] = ["GENCONID"]
        result = query_wrappers.most_recent_records_before_start_time(
            self.dummyGenConData2, start_time, table_name
        ).reset_index(drop=True)
        aim = pd.DataFrame(
            {
                "EFFECTIVEDATE": ["2017/01/04 00:15:00", "2017/01/04 00:15:00"],
                "VERSIONNO": ["1", "2"],
                "GENCONID": ["ID1", "ID1"],
            }
        )
        aim["EFFECTIVEDATE"] = pd.to_datetime(
            aim["EFFECTIVEDATE"], format="%Y/%m/%d %H:%M:%S"
        )
        assert_frame_equal(aim, result)

    def test_2_id_and_repeated_effectivedate_start_date_after_all(self):
        start_time = datetime.strptime("2019/06/01 00:00:00", "%Y/%m/%d %H:%M:%S")
        table_name = "dummy"
        defaults.primary_date_columns["dummy"] = "EFFECTIVEDATE"
        defaults.effective_date_group_col["dummy"] = ["GENCONID"]
        result = query_wrappers.most_recent_records_before_start_time(
            self.dummyGenConData3, start_time, table_name
        ).reset_index(drop=True)
        aim = pd.DataFrame(
            {
                "EFFECTIVEDATE": [
                    "2017/01/04 00:15:00",
                    "2017/01/04 00:15:00",
                    "2017/01/04 00:15:00",
                    "2017/01/04 00:15:00",
                ],
                "VERSIONNO": ["1", "2", "1", "2"],
                "GENCONID": ["ID1", "ID1", "ID2", "ID2"],
            }
        )
        aim["EFFECTIVEDATE"] = pd.to_datetime(
            aim["EFFECTIVEDATE"], format="%Y/%m/%d %H:%M:%S"
        )
        assert_frame_equal(aim, result)



import pandas as pd

from nemosis import defaults as _defaults


def _parse_datetime(series):
    """
    Attempts to parse a column into a datetime
    If unable to (because the data is not a datetime), will raise a ValueError
    Args:
        series: a numpy array (pandas column)

    Returns:
        series (np.Array)
    """

    try:
        # this first format is the most common
        return pd.to_datetime(series, format=_defaults.date_formats[0])
    except ValueError as e:
        try:
            # this format with milliseconds is used in some bidding columns
            return pd.to_datetime(series, format=_defaults.date_formats[1])
        except ValueError as e:
            # this format is used in some 4-second FCAS data
            return pd.to_datetime(series, format=_defaults.date_formats[2])


def _parse_column(series):
    """
    Attempts to parse a column into a datetime or numeric.
    If unable to, returns the original column
    Args:
        series: a numpy array (pandas column)

    Returns:
        series (np.Array)
    """

    try:
        return _parse_datetime(series)
    except ValueError:
        try:
            col_new = pd.to_numeric(series)
            return col_new
        except ValueError as e:
            return series


def _infer_column_data_types(data):
    """
    Infer datatype of DataFrame assuming inference need only be carried out
    for any columns with dtype "object". Adapted from StackOverflow.

    If the column is an object type, attempt conversions to (in order of):
    1. datetime
    2. numeric

    Returns: Data with inferred types.
    """

    for col in data:
        data[col] = _parse_column(data[col])

    return data



import unittest
from datetime import datetime
from src.nemosis import date_generators


class TestYearAndMonthGen(unittest.TestCase):
    def setUp(self):
        pass

    def test_two_times_not_at_edge_of_month_return_one_month(self):
        start_time = datetime.strptime("2017/01/02 00:00:00", "%Y/%m/%d %H:%M:%S")
        end_time = datetime.strptime("2017/01/03 00:00:00", "%Y/%m/%d %H:%M:%S")
        gen = date_generators.year_and_month_gen(start_time, end_time)
        times = [(year, month, day, index) for year, month, day, index in gen]
        self.assertEqual(times[0][0], "2017")
        self.assertEqual(times[0][1], "01")
        self.assertEqual(times[0][2], None)
        self.assertEqual(times[0][3], None)
        self.assertEqual(len(times), 1)

    def test_two_times_first_at_edge_of_month_return_month_before_and_month_of_times(
        self,
    ):
        start_time = datetime.strptime("2017/02/01 00:00:00", "%Y/%m/%d %H:%M:%S")
        end_time = datetime.strptime("2017/02/03 00:00:00", "%Y/%m/%d %H:%M:%S")
        gen = date_generators.year_and_month_gen(start_time, end_time)
        times = [(year, month, day, index) for year, month, day, index in gen]
        self.assertEqual(times[0][0], "2017")
        self.assertEqual(times[0][1], "01")
        self.assertEqual(times[0][2], None)
        self.assertEqual(times[0][3], None)
        self.assertEqual(times[1][0], "2017")
        self.assertEqual(times[1][1], "02")
        self.assertEqual(times[1][2], None)
        self.assertEqual(times[1][3], None)

    def test_two_times_first_at_edge_of_year_return_month_before_and_month_of_times(
        self,
    ):
        start_time = datetime.strptime("2017/01/01 00:00:00", "%Y/%m/%d %H:%M:%S")
        end_time = datetime.strptime("2017/01/03 00:00:00", "%Y/%m/%d %H:%M:%S")
        gen = date_generators.year_and_month_gen(start_time, end_time)
        times = [(year, month, day, index) for year, month, day, index in gen]
        self.assertEqual(times[0][0], "2016")
        self.assertEqual(times[0][1], "12")
        self.assertEqual(times[0][2], None)
        self.assertEqual(times[0][3], None)
        self.assertEqual(times[1][0], "2017")
        self.assertEqual(times[1][1], "01")
        self.assertEqual(times[1][2], None)
        self.assertEqual(times[1][3], None)

    def test_two_times_second_at_edge_of_month_returns_one_month(self):
        start_time = datetime.strptime("2017/01/05 00:00:00", "%Y/%m/%d %H:%M:%S")
        end_time = datetime.strptime("2017/01/31 00:00:00", "%Y/%m/%d %H:%M:%S")
        gen = date_generators.year_and_month_gen(start_time, end_time)
        times = [(year, month, day, index) for year, month, day, index in gen]
        self.assertEqual(times[0][0], "2017")
        self.assertEqual(times[0][1], "01")
        self.assertEqual(times[0][2], None)
        self.assertEqual(times[0][3], None)
        self.assertEqual(len(times), 1)

    def test_two_times_second_at_edge_of_year_returns_one_month(self):
        start_time = datetime.strptime("2017/12/02 00:00:00", "%Y/%m/%d %H:%M:%S")
        end_time = datetime.strptime("2017/12/31 00:00:00", "%Y/%m/%d %H:%M:%S")
        gen = date_generators.year_and_month_gen(start_time, end_time)
        times = [(year, month, day, index) for year, month, day, index in gen]
        self.assertEqual(times[0][0], "2017")
        self.assertEqual(times[0][1], "12")
        self.assertEqual(times[0][2], None)
        self.assertEqual(times[0][3], None)
        self.assertEqual(len(times), 1)

    def test_two_times_in_middle_of_jan_and_march_return_3_months(self):
        start_time = datetime.strptime("2017/01/05 00:00:00", "%Y/%m/%d %H:%M:%S")
        end_time = datetime.strptime("2017/03/24 00:00:00", "%Y/%m/%d %H:%M:%S")
        gen = date_generators.year_and_month_gen(start_time, end_time)
        times = [(year, month, day, index) for year, month, day, index in gen]
        self.assertEqual(times[0][0], "2017")
        self.assertEqual(times[0][1], "01")
        self.assertEqual(times[0][2], None)
        self.assertEqual(times[0][3], None)
        self.assertEqual(times[1][0], "2017")
        self.assertEqual(times[1][1], "02")
        self.assertEqual(times[1][2], None)
        self.assertEqual(times[1][3], None)
        self.assertEqual(times[2][0], "2017")
        self.assertEqual(times[2][1], "03")
        self.assertEqual(times[2][2], None)
        self.assertEqual(times[2][3], None)
        self.assertEqual(len(times), 3)


class TestBidTableGen(unittest.TestCase):
    def setUp(self):
        pass

    def test_two_times_not_at_edge_of_month_return_one_month(self):
        start_time = datetime.strptime("2017/01/02 00:00:00", "%Y/%m/%d %H:%M:%S")
        end_time = datetime.strptime("2017/01/03 00:00:00", "%Y/%m/%d %H:%M:%S")
        gen = date_generators.bid_table_gen(start_time, end_time)
        times = [(year, month, day, index) for year, month, day, index in gen]
        self.assertEqual(times[0][0], "2017")
        self.assertEqual(times[0][1], "01")
        self.assertEqual(times[0][2], None)
        self.assertEqual(times[0][3], None)
        self.assertEqual(len(times), 1)

    def test_two_times_first_at_edge_of_month_return_month_before_and_month_of_times(
        self,
    ):
        start_time = datetime.strptime("2017/02/01 00:00:00", "%Y/%m/%d %H:%M:%S")
        end_time = datetime.strptime("2017/02/03 00:00:00", "%Y/%m/%d %H:%M:%S")
        gen = date_generators.bid_table_gen(start_time, end_time)
        times = [(year, month, day, index) for year, month, day, index in gen]
        self.assertEqual(times[0][0], "2017")
        self.assertEqual(times[0][1], "01")
        self.assertEqual(times[0][2], None)
        self.assertEqual(times[0][3], None)
        self.assertEqual(times[1][0], "2017")
        self.assertEqual(times[1][1], "02")
        self.assertEqual(times[1][2], None)
        self.assertEqual(times[1][3], None)

    def test_two_times_first_at_edge_of_year_return_month_before_and_month_of_times(
        self,
    ):
        start_time = datetime.strptime("2017/01/01 00:00:00", "%Y/%m/%d %H:%M:%S")
        end_time = datetime.strptime("2017/01/03 00:00:00", "%Y/%m/%d %H:%M:%S")
        gen = date_generators.bid_table_gen(start_time, end_time)
        times = [(year, month, day, index) for year, month, day, index in gen]
        self.assertEqual(times[0][0], "2016")
        self.assertEqual(times[0][1], "12")
        self.assertEqual(times[0][2], None)
        self.assertEqual(times[0][3], None)
        self.assertEqual(times[1][0], "2017")
        self.assertEqual(times[1][1], "01")
        self.assertEqual(times[1][2], None)
        self.assertEqual(times[1][3], None)

    def test_two_times_second_at_edge_of_month_returns_one_month(self):
        start_time = datetime.strptime("2017/01/05 00:00:00", "%Y/%m/%d %H:%M:%S")
        end_time = datetime.strptime("2017/01/31 00:00:00", "%Y/%m/%d %H:%M:%S")
        gen = date_generators.bid_table_gen(start_time, end_time)
        times = [(year, month, day, index) for year, month, day, index in gen]
        self.assertEqual(times[0][0], "2017")
        self.assertEqual(times[0][1], "01")
        self.assertEqual(times[0][2], None)
        self.assertEqual(times[0][3], None)
        self.assertEqual(len(times), 1)

    def test_two_times_second_at_edge_of_year_returns_one_month(self):
        start_time = datetime.strptime("2017/12/02 00:00:00", "%Y/%m/%d %H:%M:%S")
        end_time = datetime.strptime("2017/12/31 00:00:00", "%Y/%m/%d %H:%M:%S")
        gen = date_generators.bid_table_gen(start_time, end_time)
        times = [(year, month, day, index) for year, month, day, index in gen]
        self.assertEqual(times[0][0], "2017")
        self.assertEqual(times[0][1], "12")
        self.assertEqual(times[0][2], None)
        self.assertEqual(times[0][3], None)
        self.assertEqual(len(times), 1)

    def test_two_times_in_middle_of_jan_and_march_return_3_months(self):
        start_time = datetime.strptime("2017/01/05 00:00:00", "%Y/%m/%d %H:%M:%S")
        end_time = datetime.strptime("2017/03/24 00:00:00", "%Y/%m/%d %H:%M:%S")
        gen = date_generators.bid_table_gen(start_time, end_time)
        times = [(year, month, day, index) for year, month, day, index in gen]
        self.assertEqual(times[0][0], "2017")
        self.assertEqual(times[0][1], "01")
        self.assertEqual(times[0][2], None)
        self.assertEqual(times[0][3], None)
        self.assertEqual(times[1][0], "2017")
        self.assertEqual(times[1][1], "02")
        self.assertEqual(times[1][2], None)
        self.assertEqual(times[1][3], None)
        self.assertEqual(times[2][0], "2017")
        self.assertEqual(times[2][1], "03")
        self.assertEqual(times[2][2], None)
        self.assertEqual(times[2][3], None)
        self.assertEqual(len(times), 3)

    def test_change_from_months_to_days(self):
        start_time = datetime.strptime("2021/02/01 00:00:00", "%Y/%m/%d %H:%M:%S")
        end_time = datetime.strptime("2021/04/03 00:00:00", "%Y/%m/%d %H:%M:%S")
        gen = date_generators.bid_table_gen(start_time, end_time)
        times = [(year, month, day, index) for year, month, day, index in gen]
        self.assertEqual(times[0][0], "2021")
        self.assertEqual(times[0][1], "01")
        self.assertEqual(times[0][2], None)
        self.assertEqual(times[0][3], None)
        self.assertEqual(times[1][0], "2021")
        self.assertEqual(times[1][1], "02")
        self.assertEqual(times[1][2], None)
        self.assertEqual(times[1][3], None)
        self.assertEqual(times[2][0], "2021")
        self.assertEqual(times[2][1], "03")
        self.assertEqual(times[2][2], None)
        self.assertEqual(times[2][3], None)
        self.assertEqual(times[3][0], "2021")
        self.assertEqual(times[3][1], "04")
        self.assertEqual(times[3][2], "01")
        self.assertEqual(times[3][3], None)
        self.assertEqual(times[4][0], "2021")
        self.assertEqual(times[4][1], "04")
        self.assertEqual(times[4][2], "02")
        self.assertEqual(times[4][3], None)
        self.assertEqual(times[5][0], "2021")
        self.assertEqual(times[5][1], "04")
        self.assertEqual(times[5][2], "03")
        self.assertEqual(times[5][3], None)
        self.assertEqual(len(times), 6)

    def test_include_previous_day_if_1st_market_day_of_month(self):
        start_time = datetime.strptime("2021/05/01 05:00:00", "%Y/%m/%d %H:%M:%S")
        end_time = datetime.strptime("2021/05/03 05:00:00", "%Y/%m/%d %H:%M:%S")
        gen = date_generators.bid_table_gen(start_time, end_time)
        times = [(year, month, day, index) for year, month, day, index in gen]
        self.assertEqual(times[0][0], "2021")
        self.assertEqual(times[0][1], "04")
        self.assertEqual(times[0][2], "30")
        self.assertEqual(times[0][3], None)
        self.assertEqual(times[1][0], "2021")
        self.assertEqual(times[1][1], "05")
        self.assertEqual(times[1][2], "01")
        self.assertEqual(times[1][3], None)
        self.assertEqual(times[2][0], "2021")
        self.assertEqual(times[2][1], "05")
        self.assertEqual(times[2][2], "02")
        self.assertEqual(times[2][3], None)
        self.assertEqual(times[3][0], "2021")
        self.assertEqual(times[3][1], "05")
        self.assertEqual(times[3][2], "03")
        self.assertEqual(times[3][3], None)
        self.assertEqual(len(times), 4)

    def test_include_previous_day_if_not_1st_market_day_of_month(
        self,
    ):
        start_time = datetime.strptime("2021/05/02 05:00:00", "%Y/%m/%d %H:%M:%S")
        end_time = datetime.strptime("2021/05/03 05:00:00", "%Y/%m/%d %H:%M:%S")
        gen = date_generators.bid_table_gen(start_time, end_time)
        times = [(year, month, day, index) for year, month, day, index in gen]
        self.assertEqual(times[0][0], "2021")
        self.assertEqual(times[0][1], "05")
        self.assertEqual(times[0][2], "01")
        self.assertEqual(times[0][3], None)
        self.assertEqual(times[1][0], "2021")
        self.assertEqual(times[1][1], "05")
        self.assertEqual(times[1][2], "02")
        self.assertEqual(times[1][3], None)
        self.assertEqual(times[2][0], "2021")
        self.assertEqual(times[2][1], "05")
        self.assertEqual(times[2][2], "03")
        self.assertEqual(times[2][3], None)
        self.assertEqual(len(times), 3)


class TestCurrentTableGen(unittest.TestCase):
    def setUp(self):
        pass

    def test_include_previous_day_if_1st_market_day_of_year(self):
        start_time = datetime.strptime("2021/01/01 05:00:00", "%Y/%m/%d %H:%M:%S")
        end_time = datetime.strptime("2021/01/03 05:00:00", "%Y/%m/%d %H:%M:%S")
        gen = date_generators.current_gen(start_time, end_time)
        times = [(year, month, day, index) for year, month, day, index in gen]
        self.assertEqual(times[0][0], "2020")
        self.assertEqual(times[0][1], "12")
        self.assertEqual(times[0][2], "31")
        self.assertEqual(times[0][3], None)
        self.assertEqual(times[1][0], "2021")
        self.assertEqual(times[1][1], "01")
        self.assertEqual(times[1][2], "01")
        self.assertEqual(times[1][3], None)
        self.assertEqual(times[2][0], "2021")
        self.assertEqual(times[2][1], "01")
        self.assertEqual(times[2][2], "02")
        self.assertEqual(times[2][3], None)
        self.assertEqual(times[3][0], "2021")
        self.assertEqual(times[3][1], "01")
        self.assertEqual(times[3][2], "03")
        self.assertEqual(times[3][3], None)
        self.assertEqual(len(times), 4)

    def test_include_previous_day_if_1st_market_day_of_month(self):
        start_time = datetime.strptime("2021/05/01 05:00:00", "%Y/%m/%d %H:%M:%S")
        end_time = datetime.strptime("2021/05/03 05:00:00", "%Y/%m/%d %H:%M:%S")
        gen = date_generators.current_gen(start_time, end_time)
        times = [(year, month, day, index) for year, month, day, index in gen]
        self.assertEqual(times[0][0], "2021")
        self.assertEqual(times[0][1], "04")
        self.assertEqual(times[0][2], "30")
        self.assertEqual(times[0][3], None)
        self.assertEqual(times[1][0], "2021")
        self.assertEqual(times[1][1], "05")
        self.assertEqual(times[1][2], "01")
        self.assertEqual(times[1][3], None)
        self.assertEqual(times[2][0], "2021")
        self.assertEqual(times[2][1], "05")
        self.assertEqual(times[2][2], "02")
        self.assertEqual(times[2][3], None)
        self.assertEqual(times[3][0], "2021")
        self.assertEqual(times[3][1], "05")
        self.assertEqual(times[3][2], "03")
        self.assertEqual(times[3][3], None)
        self.assertEqual(len(times), 4)

    def test_include_previous_day_if_not_1st_market_day_of_month(
        self,
    ):
        start_time = datetime.strptime("2021/05/02 05:00:00", "%Y/%m/%d %H:%M:%S")
        end_time = datetime.strptime("2021/05/03 05:00:00", "%Y/%m/%d %H:%M:%S")
        gen = date_generators.current_gen(start_time, end_time)
        times = [(year, month, day, index) for year, month, day, index in gen]
        self.assertEqual(times[0][0], "2021")
        self.assertEqual(times[0][1], "05")
        self.assertEqual(times[0][2], "01")
        self.assertEqual(times[0][3], None)
        self.assertEqual(times[1][0], "2021")
        self.assertEqual(times[1][1], "05")
        self.assertEqual(times[1][2], "02")
        self.assertEqual(times[1][3], None)
        self.assertEqual(times[2][0], "2021")
        self.assertEqual(times[2][1], "05")
        self.assertEqual(times[2][2], "03")
        self.assertEqual(times[2][3], None)
        self.assertEqual(len(times), 3)

    def test_include_previous_day_if_before_4am(
        self,
    ):
        start_time = datetime.strptime("2022/11/01 00:00:00", "%Y/%m/%d %H:%M:%S")
        end_time = datetime.strptime("2022/11/01 05:15:00", "%Y/%m/%d %H:%M:%S")
        gen = date_generators.current_gen(start_time, end_time)
        times = [(year, month, day, index) for year, month, day, index in gen]
        self.assertEqual(times[0][0], "2022")
        self.assertEqual(times[0][1], "10")
        self.assertEqual(times[0][2], "31")
        self.assertEqual(times[0][3], None)
        self.assertEqual(times[1][0], "2022")
        self.assertEqual(times[1][1], "11")
        self.assertEqual(times[1][2], "01")
        self.assertEqual(times[1][3], None)
        self.assertEqual(len(times), 2)


class TestYearMonthDayIndexGen(unittest.TestCase):
    def setUp(self):
        pass

    def test_two_times_in_middle_of_adjacent_hours(self):
        # Whole of each hour should be returned.
        start_time = datetime.strptime("2013/01/05 12:20:00", "%Y/%m/%d %H:%M:%S")
        end_time = datetime.strptime("2013/01/05 13:45:00", "%Y/%m/%d %H:%M:%S")
        gen = date_generators.year_month_day_index_gen(start_time, end_time)
        times = [(year, month, day, index) for year, month, day, index in gen]
        self.assertEqual(times[0][0], "2013")
        self.assertEqual(times[0][1], "01")
        self.assertEqual(times[0][2], "05")
        self.assertEqual(times[0][3], "1355")
        self.assertEqual(times[-1][0], "2013")
        self.assertEqual(times[-1][1], "01")
        self.assertEqual(times[-1][2], "05")
        self.assertEqual(times[-1][3], "1200")
        self.assertEqual(len(times), 24)

    def test_two_times_one_at_start_of_year_should_not_over_flow_to_previous_year(self):
        start_time = datetime.strptime("2013/01/01 00:00:00", "%Y/%m/%d %H:%M:%S")
        end_time = datetime.strptime("2013/01/01 01:00:00", "%Y/%m/%d %H:%M:%S")
        gen = date_generators.year_month_day_index_gen(start_time, end_time)
        times = [(year, month, day, index) for year, month, day, index in gen]
        self.assertEqual(times[0][0], "2013")
        self.assertEqual(times[0][1], "01")
        self.assertEqual(times[0][2], "01")
        self.assertEqual(times[0][3], "0155")
        self.assertEqual(times[-1][0], "2013")
        self.assertEqual(times[-1][1], "01")
        self.assertEqual(times[-1][2], "01")
        self.assertEqual(times[-1][3], "0000")
        self.assertEqual(len(times), 24)

    def test_two_times_one_at_start_of_month_should_not_over_flow_to_previous_month(
        self,
    ):
        start_time = datetime.strptime("2013/02/01 00:00:00", "%Y/%m/%d %H:%M:%S")
        end_time = datetime.strptime("2013/02/01 01:00:00", "%Y/%m/%d %H:%M:%S")
        gen = date_generators.year_month_day_index_gen(start_time, end_time)
        times = [(year, month, day, index) for year, month, day, index in gen]
        self.assertEqual(times[0][0], "2013")
        self.assertEqual(times[0][1], "02")
        self.assertEqual(times[0][2], "01")
        self.assertEqual(times[0][3], "0155")
        self.assertEqual(times[-1][0], "2013")
        self.assertEqual(times[-1][1], "02")
        self.assertEqual(times[-1][2], "01")
        self.assertEqual(times[-1][3], "0000")
        self.assertEqual(len(times), 24)

    def test_no_missing_values_in_a_week(self):
        start_time = datetime.strptime("2013/02/01 00:00:00", "%Y/%m/%d %H:%M:%S")
        end_time = datetime.strptime("2013/02/07 00:00:00", "%Y/%m/%d %H:%M:%S")
        gen = date_generators.year_month_day_index_gen(start_time, end_time)
        times = [(year, month, day, index) for year, month, day, index in gen]
        self.assertEqual(times[0][0], "2013")
        self.assertEqual(times[0][1], "02")
        self.assertEqual(times[0][2], "01")
        self.assertEqual(times[0][3], "2355")
        self.assertEqual(times[-1][0], "2013")
        self.assertEqual(times[-1][1], "02")
        self.assertEqual(times[-1][2], "07")
        self.assertEqual(times[-1][3], "0000")
        self.assertEqual(len(times), 1740)



import logging
import os as _os
import glob as _glob
import pandas as _pd
from datetime import datetime as _datetime
from nemosis import filters as _filters
from nemosis import downloader as _downloader
from nemosis import processing_info_maps as _processing_info_maps
from nemosis import defaults as _defaults
from nemosis import custom_tables as _custom_tables
from nemosis import _infer_column_data_types
from nemosis.custom_errors import UserInputError, NoDataToReturn, DataMismatchError

logger = logging.getLogger(__name__)


def dynamic_data_compiler(
    start_time,
    end_time,
    table_name,
    raw_data_location,
    select_columns=None,
    filter_cols=None,
    filter_values=None,
    fformat="feather",
    keep_csv=True,
    parse_data_types=True,
    rebuild=False,
    **kwargs,
):
    """
    Downloads and compiles data for all dynamic tables. For non-CSV formats,
    will save data typed as strings/objects. To save typed data (e.g.
    appropriate cols are Float or Int), use cache_compiler.
    Args:
        start_time (str): format 'yyyy/mm/dd HH:MM:SS'.
        end_time (str): format 'yyyy/mm/dd HH:MM:SS'.
        table_name (str): table as per Wiki.
        raw_data_location (str): directory to download and cache data to.
                                 existing data will be used if in this dir.
        select_columns (list or str): a list of columns to return or the string
                                      'all' to return all columns from AMEO raw
                                      data, 'all argument must be used will
                                      fformat='csv'. Default is None, will return a
                                      default set of columns.
        filter_cols (list): filter on columns.
        filter_values (tuple[list]): filter index n filter col such that values are
                              equal to index n filter value.
        fformat (string): "csv", "feather" or "parquet" for storage and access.
                          Stored parquet and feather files will store columns
                          as object type (compatbile with GUI use). For
                          type inference for a cache, use cache_compiler.
        keep_csv (bool): retains CSVs in cache.
        data_merge (bool): concatenate DataFrames and return one DataFrame.
                           If False, will not return any data.
        parse_data_types (bool): infers data types of columns when reading
                                 data. default True for API use.
        rebuild (bool): If True then cache files are rebuilt
                        even if they exist already. False by default.
        **kwargs: additional arguments passed to the pd.to_{fformat}() function

    Returns:
        all_data (pd.Dataframe): All data concatenated.
    """
    if not _os.path.isdir(raw_data_location):
        raise UserInputError("The raw_data_location provided does not exist.")

    if table_name not in _defaults.dynamic_tables:
        raise UserInputError("Table name provided is not a dynamic table.")

    if fformat not in ["csv", "feather", "parquet"]:
        raise UserInputError("Argument fformat must be 'csv', 'feather' or 'parquet'")

    if select_columns == "all" and fformat != "csv":
        raise UserInputError(
            "If select_columns='all' is used fformat='csv' must be used."
        )

    (
        start_time,
        end_time,
        select_columns,
        date_filter,
        start_search,
    ) = _set_up_dynamic_compilers(table_name, start_time, end_time, select_columns)

    if filter_cols and not set(filter_cols).issubset(set(select_columns)):
        raise UserInputError(
            (
                "Filter columns not valid. They must be a part of "
                + "select_columns or the table defaults."
            )
        )

    logger.info(f"Compiling data for table {table_name}")

    start_time = _datetime.strptime(start_time, "%Y/%m/%d %H:%M:%S")
    end_time = _datetime.strptime(end_time, "%Y/%m/%d %H:%M:%S")
    start_search = _datetime.strptime(start_search, "%Y/%m/%d %H:%M:%S")
    data_tables = _dynamic_data_fetch_loop(
        start_search,
        start_time,
        end_time,
        table_name,
        raw_data_location,
        select_columns,
        date_filter,
        fformat=fformat,
        keep_csv=keep_csv,
        rebuild=rebuild,
        write_kwargs=kwargs,
    )
    if data_tables:
        all_data = _pd.concat(data_tables, sort=False)
        finalise_data = _processing_info_maps.finalise[table_name]
        if finalise_data is not None:
            for function in finalise_data:
                all_data = function(all_data, start_time, table_name)

        if parse_data_types:
            all_data = _infer_column_data_types(all_data)
        if filter_cols is not None:
            if not set(filter_cols).issubset(set(all_data.columns)):
                missing_columns = [
                    col for col in filter_cols if col not in all_data.columns
                ]
                UserInputError(f"Filter columns {missing_columns} not in data.")
            else:
                all_data = _filters.filter_on_column_value(
                    all_data, filter_cols, filter_values
                )
        logger.info(f"Returning {table_name}.")
        return all_data
    else:
        raise NoDataToReturn(
            (
                f"Compiling data for table {table_name} failed. "
                + "This probably because none of the requested data "
                + "could be download from AEMO. Check your internet "
                + "connection and that the requested data is archived on: "
                + "https://nemweb.com.au see nemosis.defaults for table specific urls."
            )
        )


def cache_compiler(
    start_time,
    end_time,
    table_name,
    raw_data_location,
    select_columns=None,
    fformat="feather",
    rebuild=False,
    keep_csv=False,
    **kwargs,
):
    """
    Downloads and compiles typed data for all dynamic tables as either parquet
    or feather format (i.e. will save data with columns as appropriate data
    types such as Int, Float or Datetime=False).

    Should not be used in a cache
    that is used to store csvs (such as the cache for the GUI).

    Args:
        start_time (str): format 'yyyy/mm/dd HH:MM:SS'.
        end_time (str): format 'yyyy/mm/dd HH:MM:SS'.
        table_name (str): table as per Wiki.
        raw_data_location (str): directory to download and cache data to.
                                 existing data will be used if in this dir.
        select_columns (list or str): a list of columns to return, or the string
                                     'all' to return all columns from AMEO raw
                                     data. Determines which columns are included in cache file.
                                     Default is None, will return a default set of columns.
        fformat (string): "feather" or "parquet" for storage and access.
                          Stored parquet and feather files will store columns
                          as object type (compatbile with GUI use). For
                          type inference for a cache, use cache_compiler.
        rebuild (bool): If True then cache files are rebuilt
                        even if they exist already. False by default.
        keep_csv (bool): If True raw CSVs from AEMO are not deleted after
                         the cache is built. False by default
        **kwargs: additional arguments passed to the pd.to_{fformat}() function

    Returns:
        Nothing
    """
    if not _os.path.isdir(raw_data_location):
        raise UserInputError("The raw_data_location provided does not exist.")

    if table_name not in _defaults.dynamic_tables:
        raise UserInputError("Table name provided is not a dynamic table.")

    if fformat != "parquet" and fformat != "feather":
        raise UserInputError("Argument fformat must be 'feather' or 'parquet'")

    if select_columns is not None and not rebuild:
        raise UserInputError(
            (
                "The select_columns argument must be used with rebuild=True "
                + "to ensure the cache is built with the correct columns."
            )
        )

    logger.info(f"Caching data for table {table_name}")

    (
        start_time,
        end_time,
        select_columns,
        _,
        start_search,
    ) = _set_up_dynamic_compilers(table_name, start_time, end_time, select_columns)
    start_time = _datetime.strptime(start_time, "%Y/%m/%d %H:%M:%S")
    end_time = _datetime.strptime(end_time, "%Y/%m/%d %H:%M:%S")
    start_search = _datetime.strptime(start_search, "%Y/%m/%d %H:%M:%S")

    _dynamic_data_fetch_loop(
        start_search,
        start_time,
        end_time,
        table_name,
        raw_data_location,
        select_columns,
        date_filter=None,
        fformat=fformat,
        keep_csv=keep_csv,
        caching_mode=True,
        rebuild=rebuild,
        write_kwargs=kwargs,
    )
    return


def static_table(
    table_name,
    raw_data_location,
    select_columns=None,
    filter_cols=None,
    filter_values=None,
    update_static_file=False,
):
    """
    Downloads and compiles data for all static tables.
    Args:
        table_name (str): table as per Wiki.
        raw_data_location (str): directory to download and cache data to.
                                 existing data will be used if in this dir.
        select_columns (list or str): a list of columns to return or the string
                                      'all' to return all columns from AMEO raw
                                      data, 'all argument must be used will
                                      fformat='csv'. Default is None, will return a
                                      default set of columns.
        filter_cols (list): filter on columns.
        filter_values (tuple[list]): filter index n filter col such that values are
                              equal to index n filter value.
        update_static_file (bool): If True download latest version of file
                                   even if a version already exists.
                                   Default is False.

    Returns:
        data (pd.Dataframe)
    """
    if not _os.path.isdir(raw_data_location):
        raise UserInputError("The raw_data_location provided does not exist.")

    if table_name not in _defaults.static_tables:
        raise UserInputError("Table name provided is not a static table.")

    if filter_cols and not set(filter_cols).issubset(set(select_columns)):
        raise UserInputError(
            (
                "Filter columns not valid. They must be a part of "
                + "select_columns or the table defaults."
            )
        )

    logger.info(f"Retrieving static table {table_name}")
    path_and_name = _os.path.join(raw_data_location, _defaults.names[table_name])
    if not _os.path.isfile(path_and_name) or update_static_file:
        logger.info(f"Downloading data for table {table_name}")
        try:
            static_downloader_map[table_name](
                _defaults.static_table_url[table_name], path_and_name
            )
        except:
            raise NoDataToReturn(
                (
                    f"Compiling data for table {table_name} failed. "
                    + "This probably because none of the requested data "
                    + "could be download from AEMO. Check your internet "
                    + "connection and that the requested data is archived on: "
                    + "https://nemweb.com.au see nemosis.defaults for table specific urls."
                )
            )

    table = static_file_reader_map[table_name](path_and_name, table_name)

    if select_columns != "all":

        if select_columns is None:
            select_columns = _defaults.table_columns[table_name]

        read_cols = _validate_select_columns(table, select_columns, path_and_name)

        if not read_cols:
            raise DataMismatchError(
                (
                    f"None of columns {select_columns} are in {path_and_name}. "
                    "This may be caused by user input if the 'select_columns' "
                    "argument is being used, or by changed AEMO data formats. "
                    "This error can be avoided by using the argument select_columns='all'."
                )
            )

        table = table.loc[:, read_cols]

    for column in table.select_dtypes(["object"]).columns:
        table[column] = table[column].map(lambda x: _strip_if_string(x))

    if filter_cols is not None:
        if not set(filter_cols).issubset(set(table.columns)):
            missing_columns = [col for col in filter_cols if col not in table.columns]
            UserInputError(f"Filter columns {missing_columns} not in data.")
        else:
            table = _filters.filter_on_column_value(table, filter_cols, filter_values)

    static_table_finalisers = static_data_finaliser_map[table_name]
    for finaliser in static_table_finalisers:
        table = finaliser(table, table_name)

    return table


def _strip_if_string(x):
    if isinstance(x, str):
        x = x.strip()
    return x


def _get_read_function(fformat, table_type, day):
    if fformat == "feather":
        func = _pd.read_feather
    elif fformat == "parquet":
        func = _pd.read_parquet
    elif fformat == "csv":
        if table_type == "MMS":
            func = _read_mms_csv
        elif table_type == "FCAS":
            func = _read_fcas_causer_pays_csv
        elif table_type == "BIDDING":
            if day is None:
                func = _read_mms_csv
            else:
                func = _read_constructed_csv
        elif table_type in ['DAILY_REGION_SUMMARY', "NEXT_DAY_DISPATCHLOAD", "INTERMITTENT_GEN_SCADA"]:
            func = _read_constructed_csv
    return func


def _count_csv_lines(file_path):
    with open(file_path, 'rb') as f:
        return sum(1 for _ in f)


def _read_mms_csv(path_and_name, dtype=None, usecols=None, nrows=None, names=None):
    last_line_number = _count_csv_lines(path_and_name) - 1
    data = _pd.read_csv(
        path_and_name,
        skiprows=[0, last_line_number],
        dtype=dtype,
        usecols=usecols,
        nrows=nrows,
        names=names,
    )
    return data


def _read_constructed_csv(
    path_and_name, dtype=None, usecols=None, nrows=None, names=None
):
    data = _pd.read_csv(
        path_and_name, dtype=dtype, usecols=usecols, nrows=nrows, names=names
    )
    return data


def _read_fcas_causer_pays_csv(
    path_and_name, dtype=None, usecols=None, nrows=None, names=None
):
    data = _pd.read_csv(
        path_and_name, dtype=dtype, usecols=usecols, nrows=nrows, names=names
    )
    return data


def _read_static_csv(path_and_name, table_name):
    return _pd.read_csv(
        path_and_name, dtype=str, names=_defaults.table_columns[table_name]
    )


def _read_excel(path_and_name, table_name):
    xls = _pd.ExcelFile(path_and_name)
    for tab_option in _defaults.reg_exemption_list_tabs[table_name]:
        try:
            return _pd.read_excel(xls, tab_option, dtype=str)
        except ValueError as e:
            pass
    raise NoDataToReturn(f"""
        The excel file did not have any of the expected tabs {_defaults.reg_exemption_list_tabs[table_name]}.
        """)


def _finalise_excel_data(data, table_name):
    if table_name in _defaults.table_primary_keys.keys():
        primary_keys = _defaults.table_primary_keys[table_name]
        data = data.drop_duplicates(primary_keys)
    data = data.dropna(axis=0, how="all")
    data = data.dropna(axis=1, how="all")
    return data


def _finalise_generators_and_scheduled_loads(data, table_name):
    data = data.replace(to_replace=['', ' '], value='-')
    data = data.fillna('-')
    return data
    

def _finalise_csv_data(data, table_name):
    return data


static_downloader_map = {
    "VARIABLES_FCAS_4_SECOND": _downloader.download_csv,
    "ELEMENTS_FCAS_4_SECOND": _downloader.download_elements_file,
    "Generators and Scheduled Loads": _downloader.download_xl,
    "_downloader.download_xl": _downloader.download_xl,
}

static_file_reader_map = {
    "VARIABLES_FCAS_4_SECOND": _read_static_csv,
    "ELEMENTS_FCAS_4_SECOND": _read_static_csv,
    "Generators and Scheduled Loads": _read_excel,
    "FCAS Providers": _read_excel,
}

static_data_finaliser_map = {
    "VARIABLES_FCAS_4_SECOND": [_finalise_csv_data],
    "ELEMENTS_FCAS_4_SECOND": [_finalise_csv_data],
    "Generators and Scheduled Loads": [_finalise_excel_data, _finalise_generators_and_scheduled_loads],
    "FCAS Providers": [_finalise_excel_data],
}


def static_table_FCAS_elements_file(
    table_name,
    raw_data_location,
    select_columns=None,
    filter_cols=None,
    filter_values=None,
    update_static_file=False,
):
    table = static_table(
        table_name,
        raw_data_location,
        select_columns,
        filter_cols,
        filter_values,
        update_static_file,
    )
    return table


def static_table_xl(
    table_name,
    raw_data_location,
    select_columns=None,
    filter_cols=None,
    filter_values=None,
    update_static_file=False,
):
    table = static_table(
        table_name,
        raw_data_location,
        select_columns,
        filter_cols,
        filter_values,
        update_static_file,
    )
    return table


def _set_up_dynamic_compilers(table_name, start_time, end_time, select_columns):
    """
    Set up function for compilers that deal with dynamic data.

    Returns: start_time, end_time, select_columns, defaults_columns,
             date_filter, start_search, search_type.
    """
    # Generic setup common to all tables.
    default_cols = _defaults.table_columns[table_name]
    if select_columns is None:
        select_columns = default_cols

    # Pre loop setup, done at table type basis.
    date_filter = _processing_info_maps.filter[table_name]
    setup_function = _processing_info_maps.setup[table_name]
    if setup_function is not None:
        start_time, end_time = setup_function(start_time, end_time)

    search_type = _processing_info_maps.search_type[table_name]

    if search_type == "all":
        start_search = _defaults.nem_data_model_start_time
    elif search_type == "start_to_end":
        start_search = start_time
    elif search_type == "end":
        start_search = end_time
    return start_time, end_time, select_columns, date_filter, start_search


def _dynamic_data_fetch_loop(
    start_search,
    start_time,
    end_time,
    table_name,
    raw_data_location,
    select_columns,
    date_filter,
    fformat="feather",
    keep_csv=True,
    caching_mode=False,
    rebuild=False,
    write_kwargs={},
):
    """
    Loops through generated dates and checks if the appropriate file exists.

    If it does, reads in the data from the file and performs filtering.

    If it does not, check if the CSV exists:
    1. If it does, read the data in and write any required files
       (parquet or feather).
    2. If it does not, download data then do the same as 1.
    """
    data_tables = []

    table_type = _defaults.table_types[table_name]
    date_gen = _processing_info_maps.date_gen[table_type](start_search, end_time)

    for year, month, day, index in date_gen:
        check_for_next_data_chunk = True
        chunk = 0
        while check_for_next_data_chunk:
            chunk += 1

            filename_stub, full_filename, path_and_name = _create_filename(
                table_name, table_type, raw_data_location, fformat, day, month, year, chunk, index
            )

            if not (
                _glob.glob(full_filename) or _glob.glob(path_and_name + ".[cC][sS][vV]")
            ) or (not _glob.glob(path_and_name + ".[cC][sS][vV]") and rebuild):
                _download_data(
                    table_name,
                    table_type,
                    filename_stub,
                    day,
                    month,
                    year,
                    chunk,
                    index,
                    raw_data_location,
                )

            if _glob.glob(full_filename) and fformat != "csv" and not rebuild:
                if not caching_mode:
                    data = _get_read_function(fformat, table_type, day)(full_filename)
                else:
                    data = None
                    logger.info(
                        f"Cache for {table_name} in date range already compiled in"
                        + f" {raw_data_location}."
                    )

            elif _glob.glob(path_and_name + ".[cC][sS][vV]"):

                if select_columns != "all":
                    read_all_columns = False
                else:
                    read_all_columns = True

                if not caching_mode:
                    dtypes = "str"
                else:
                    dtypes = "all"

                csv_path_and_name = _glob.glob(path_and_name + ".[cC][sS][vV]")[0]

                csv_read_function = _get_read_function(
                    fformat="csv", table_type=table_type, day=day
                )
                data = _determine_columns_and_read_csv(
                    table_name,
                    csv_path_and_name,
                    csv_read_function,
                    read_all_columns=read_all_columns,
                    dtypes=dtypes,
                )

                if caching_mode:
                    data = _perform_column_selection(data, select_columns, full_filename)

                if data is not None and fformat != "csv":
                    _log_file_creation_message(fformat, table_name, year, month, day, index)
                    _write_to_format(data, fformat, full_filename, write_kwargs)

                if not keep_csv:
                    _os.remove(_glob.glob(path_and_name + ".[cC][sS][vV]")[0])
            else:
                data = None

            if not caching_mode and data is not None:

                if date_filter is not None:
                    data = date_filter(data, start_time, end_time)

                data = _perform_column_selection(data, select_columns, full_filename)

                data_tables.append(data)
            elif not caching_mode and chunk == 1:
                logger.warning(f"Loading data from {full_filename} failed.")

            if data is None or '#' not in filename_stub:
                check_for_next_data_chunk = False

    return data_tables


def _perform_column_selection(data, select_columns, full_filename):
    if select_columns != "all":
        keep_cols = _validate_select_columns(data, select_columns, full_filename)
        if keep_cols:
            data = data.loc[:, keep_cols]
        else:
            raise DataMismatchError(
                (
                    f"None of columns {select_columns} are in {full_filename}. "
                    "This may be caused by user input if the 'select_columns' "
                    "argument is being used, or by changed AEMO data formats. "
                    "This error can be avoided by using the argument select_columns='all'."
                )
            )
    return data


def _create_filename(
    table_name, table_type, raw_data_location, fformat, day, month, year, chunk, index
):
    """
    Gather:
    - the file name, based on file naming rules
    - potential file path (if data exists in cache)

    Returns: filename_stub, full_filename and path_and_name
    """
    filename_stub, path_and_name = _processing_info_maps.write_filename[table_type](
        table_name, month, year, day, chunk, index, raw_data_location
    )
    full_filename = path_and_name + f".{fformat}"
    return filename_stub, full_filename, path_and_name


def _validate_select_columns(data, select_columns, full_filename):
    """
    Checks whether select_columns are in the file. If at least one is,
    then it will return any of select_columns that are available as well as
    the date col (for date filtering).  If not, it will return an empty list.

    Returns: List
    """
    file_cols = data.columns
    available_cols = file_cols[file_cols.isin(select_columns)].tolist()
    rejected_cols = set(select_columns) - set(available_cols)
    if not available_cols:
        return []
    else:
        if rejected_cols:
            logger.warning(
                f"{rejected_cols} not in {full_filename}. "
                + f"Loading {available_cols}"
            )
        return available_cols


def _log_file_creation_message(fformat, table_name, year, month, day, index):
    logstr = f"Creating {fformat} file for " + f"{table_name}, {year}, {month}"
    if day is None:
        output = logstr
    else:
        output = logstr + f" {day}, {index}"

    logger.info(output)


def _determine_columns_and_read_csv(
    table_name, csv_file, read_csv_func, dtypes, read_all_columns=False
):
    """
    Used by read_data_and_create_file
    Determining columns:
    - If the table is an MMS table, check header of CSV for actual columns.
      Then remove any columns from lookup table if not in actual columns.
      This is done as AEMO has added and removed columns over time.
    - If the table is not an MMS table, use columns from the lookup table.

    Reading csv:
    - To preserve compatability with previous versions of NEMOSIS and
      thus any existing data caches, read in all columns as strings.

    Returns: data, columns
    """
    if dtypes == "all":
        type = None
    else:
        type = str
    if (
        _defaults.table_types[table_name] in ["MMS", "BIDDING", "DAILY_REGION_SUMMARY", "NEXT_DAY_DISPATCHLOAD",
                                              "INTERMITTENT_GEN_SCADA"]
        and not read_all_columns
    ):
        headers = read_csv_func(csv_file, nrows=1).columns.tolist()
        columns = [
            column
            for column in _defaults.table_columns[table_name]
            if column in headers
        ]
        data = read_csv_func(csv_file, usecols=columns, dtype=type)
    elif (
        _defaults.table_types[table_name] in ["MMS", "BIDDING", "DAILY_REGION_SUMMARY", "NEXT_DAY_DISPATCHLOAD"]
        and read_all_columns
    ):
        data = read_csv_func(csv_file, dtype=type)
    else:
        columns = _defaults.table_columns[table_name]
        data = read_csv_func(csv_file, names=columns, dtype=type)
    return data


def _write_to_format(data, fformat, full_filename, write_kwargs):
    """
    Used by read_data_and_create_file
    Writes the DataFrame to a non-CSV format is a non_CSV format is specified.
    """
    write_function = {"feather": data.to_feather, "parquet": data.to_parquet}
    # Remove files of the same name - deals with case of corrupted files.
    if _os.path.isfile(full_filename) and fformat != "csv":
        _os.unlink(full_filename)
    # Write to required format
    if fformat == "feather":
        write_function[fformat](full_filename, **write_kwargs)
    elif fformat == "parquet":
        write_function[fformat](full_filename, index=False, **write_kwargs)
    return


def _download_data(
    table_name, table_type, filename_stub, day, month, year, chunk, index, raw_data_location
):
    """
    Dispatch table to downloader to be downloaded.

    Returns: nothing
    """
    if chunk == 1:
        if day is None:
            logger.info(
                f"Downloading data for table {table_name}, " + f"year {year}, month {month}"
            )
        elif index is None:
            logger.info(
                f"Downloading data for table {table_name}, "
                + f"year {year}, month {month}, day {day}"
            )
        else:
            logger.info(
                f"Downloading data for table {table_name}, "
                + f"year {year}, month {month}, day {day},"
                + f"time {index}."
            )

    _processing_info_maps.downloader[table_type](
        year, month, day, chunk, index, filename_stub, raw_data_location
    )
    return

  
# GUI wrappers and mappers below


def _dynamic_data_wrapper_for_gui(
    start_time, end_time, table, raw_data_location, columns, filter_cols, filter_values
):
    data = dynamic_data_compiler(
        start_time=start_time,
        end_time=end_time,
        table_name=table,
        raw_data_location=raw_data_location,
        select_columns=columns,
        filter_cols=filter_cols,
        filter_values=filter_values,
        parse_data_types=False,
    )
    return data


def _static_table_wrapper_for_gui(
    start_time,
    end_time,
    table_name,
    raw_data_location,
    select_columns=None,
    filter_cols=None,
    filter_values=None,
):
    table = static_table(
        table_name, raw_data_location, select_columns, filter_cols, filter_values
    )
    return table


_method_map = {
    "DISPATCHLOAD": _dynamic_data_wrapper_for_gui,
    "DISPATCHPRICE": _dynamic_data_wrapper_for_gui,
    "TRADINGLOAD": _dynamic_data_wrapper_for_gui,
    "TRADINGPRICE": _dynamic_data_wrapper_for_gui,
    "TRADINGREGIONSUM": _dynamic_data_wrapper_for_gui,
    "TRADINGINTERCONNECT": _dynamic_data_wrapper_for_gui,
    "DISPATCH_UNIT_SCADA": _dynamic_data_wrapper_for_gui,
    "DISPATCHCONSTRAINT": _dynamic_data_wrapper_for_gui,
    "DUDETAILSUMMARY": _dynamic_data_wrapper_for_gui,
    "PARTICIPANT": _dynamic_data_wrapper_for_gui,
    "DUDETAIL": _dynamic_data_wrapper_for_gui,
    "GENCONDATA": _dynamic_data_wrapper_for_gui,
    "SPDREGIONCONSTRAINT": _dynamic_data_wrapper_for_gui,
    "SPDCONNECTIONPOINTCONSTRAINT": _dynamic_data_wrapper_for_gui,
    "SPDINTERCONNECTORCONSTRAINT": _dynamic_data_wrapper_for_gui,
    "FCAS_4_SECOND": _dynamic_data_wrapper_for_gui,
    "ELEMENTS_FCAS_4_SECOND": _static_table_wrapper_for_gui,
    "VARIABLES_FCAS_4_SECOND": _static_table_wrapper_for_gui,
    "Generators and Scheduled Loads": _static_table_wrapper_for_gui,
    "FCAS Providers": _static_table_wrapper_for_gui,
    "BIDDAYOFFER_D": _dynamic_data_wrapper_for_gui,
    "BIDPEROFFER_D": _dynamic_data_wrapper_for_gui,
    "FCAS_4s_SCADA_MAP": _custom_tables.fcas4s_scada_match,
    "PLANTSTATS": _custom_tables.plant_stats,
    "DISPATCHINTERCONNECTORRES": _dynamic_data_wrapper_for_gui,
    "DISPATCHREGIONSUM": _dynamic_data_wrapper_for_gui,
    "LOSSMODEL": _dynamic_data_wrapper_for_gui,
    "LOSSFACTORMODEL": _dynamic_data_wrapper_for_gui,
    "MNSP_DAYOFFER": _dynamic_data_wrapper_for_gui,
    "MNSP_PEROFFER": _dynamic_data_wrapper_for_gui,
    "MNSP_INTERCONNECTOR": _dynamic_data_wrapper_for_gui,
    "INTERCONNECTOR": _dynamic_data_wrapper_for_gui,
    "INTERCONNECTORCONSTRAINT": _dynamic_data_wrapper_for_gui,
    "MARKET_PRICE_THRESHOLDS": _dynamic_data_wrapper_for_gui,
    "DAILY_REGION_SUMMARY": _dynamic_data_wrapper_for_gui,
    "NEXT_DAY_DISPATCHLOAD": _dynamic_data_wrapper_for_gui,
    "INTERMITTENT_GEN_SCADA": _dynamic_data_wrapper_for_gui,
    "ROOFTOP_PV_ACTUAL": _dynamic_data_wrapper_for_gui
}



import unittest
from src.nemosis import dynamic_data_compiler, cache_compiler, static_table
from src.nemosis import defaults
import os
from unittest.mock import patch
from io import StringIO


class TestDynamicDataCompilerRaisesExpectedErrors(unittest.TestCase):
    def test_raise_error_for_incorrect_table_name(self):
        with self.assertRaises(Exception) as context:
            dynamic_data_compiler(
                "2000/01/01 00:00:00",
                "2000/02/01 00:00:00",
                "NOTATABLE",
                defaults.raw_data_cache,
            )
        self.assertTrue(
            "Table name provided is not a dynamic table." in str(context.exception)
        )

    def test_raise_error_for_no_data_returned(self):
        with self.assertRaises(Exception) as context:
            dynamic_data_compiler(
                "2000/01/01 00:00:00",
                "2000/02/01 00:00:00",
                "DISPATCHPRICE",
                defaults.raw_data_cache,
            )
        self.assertTrue(
            (
                f"Compiling data for table DISPATCHPRICE failed. "
                + "This probably because none of the requested data "
                + "could be download from AEMO. Check your internet "
                + "connection and that the requested data is archived on: "
                + "https://nemweb.com.au see nemosis.defaults for table specific urls."
            )
            in str(context.exception)
        )

    def test_raise_error_for_filter_column_not_in_select_columns(self):
        with self.assertRaises(Exception) as context:
            dynamic_data_compiler(
                "2019/01/01 00:00:00",
                "2019/02/01 00:00:00",
                "DISPATCHPRICE",
                defaults.raw_data_cache,
                select_columns=["REGIONID", "SETTLEMENTDATE", "RRP"],
                filter_cols=["INTERVENTION"],
                filter_values=(["0"],),
            )
        self.assertTrue(
            (
                "Filter columns not valid. They must be a part of "
                + "select_columns or the table defaults."
            )
            in str(context.exception)
        )

    def test_raise_error_for_filter_column_not_in_default_columns(self):
        with self.assertRaises(Exception) as context:
            dynamic_data_compiler(
                "2019/01/01 00:00:00",
                "2019/02/01 00:00:00",
                "DISPATCHPRICE",
                defaults.raw_data_cache,
                select_columns=["REGIONID", "SETTLEMENTDATE", "RRP"],
                filter_cols=["NOTACOLUMN"],
                filter_values=(["0"],),
            )
        self.assertTrue(
            (
                "Filter columns not valid. They must be a part of "
                + "select_columns or the table defaults."
            )
            in str(context.exception)
        )

    def test_raise_error_if_fformat_not_in_expected_set(self):
        with self.assertRaises(Exception) as context:
            dynamic_data_compiler(
                "2019/01/01 00:00:00",
                "2019/02/01 00:00:00",
                "DISPATCHPRICE",
                defaults.raw_data_cache,
                fformat="db",
            )
        self.assertTrue(
            "Argument fformat must be 'csv', 'feather' or 'parquet'"
            in str(context.exception)
        )

    def test_raise_error_if_select_columns_not_in_data(self):
        with self.assertRaises(Exception) as context:
            dynamic_data_compiler(
                "2019/01/01 00:00:00",
                "2019/02/01 00:00:00",
                "DISPATCHPRICE",
                defaults.raw_data_cache,
                select_columns=["NOTACOLUMN"],
            )
        self.assertIn(
            (
                f"None of columns ['NOTACOLUMN'] are in D:/nemosis_test_cache\\PUBLIC_DVD_DISPATCHPRICE_201812010000.feather. "
                "This may be caused by user input if the 'select_columns' "
                "argument is being used, or by changed AEMO data formats. "
                "This error can be avoided by using the argument select_columns='all'."
            ),
            str(context.exception)
        )

    def test_using_select_columns_all_does_not_raise_error(self):
        price_data = dynamic_data_compiler(
            "2019/01/01 00:00:00",
            "2019/02/01 00:00:00",
            "DISPATCHPRICE",
            defaults.raw_data_cache,
            select_columns="all",
            fformat="csv",
        )
        expected_columns = [
            "I",
            "DISPATCH",
            "PRICE",
            "1",
            "SETTLEMENTDATE",
            "RUNNO",
            "REGIONID",
            "DISPATCHINTERVAL",
            "INTERVENTION",
            "RRP",
            "EEP",
            "ROP",
            "APCFLAG",
            "MARKETSUSPENDEDFLAG",
            "LASTCHANGED",
            "RAISE6SECRRP",
            "RAISE6SECROP",
            "RAISE6SECAPCFLAG",
            "RAISE60SECRRP",
            "RAISE60SECROP",
            "RAISE60SECAPCFLAG",
            "RAISE5MINRRP",
            "RAISE5MINROP",
            "RAISE5MINAPCFLAG",
            "RAISEREGRRP",
            "RAISEREGROP",
            "RAISEREGAPCFLAG",
            "LOWER6SECRRP",
            "LOWER6SECROP",
            "LOWER6SECAPCFLAG",
            "LOWER60SECRRP",
            "LOWER60SECROP",
            "LOWER60SECAPCFLAG",
            "LOWER5MINRRP",
            "LOWER5MINROP",
            "LOWER5MINAPCFLAG",
            "LOWERREGRRP",
            "LOWERREGROP",
            "LOWERREGAPCFLAG",
            "PRICE_STATUS",
            "PRE_AP_ENERGY_PRICE",
            "PRE_AP_RAISE6_PRICE",
            "PRE_AP_RAISE60_PRICE",
            "PRE_AP_RAISE5MIN_PRICE",
            "PRE_AP_RAISEREG_PRICE",
            "PRE_AP_LOWER6_PRICE",
            "PRE_AP_LOWER60_PRICE",
            "PRE_AP_LOWER5MIN_PRICE",
            "PRE_AP_LOWERREG_PRICE",
            "CUMUL_PRE_AP_ENERGY_PRICE",
            "CUMUL_PRE_AP_RAISE6_PRICE",
            "CUMUL_PRE_AP_RAISE60_PRICE",
            "CUMUL_PRE_AP_RAISE5MIN_PRICE",
            "CUMUL_PRE_AP_RAISEREG_PRICE",
            "CUMUL_PRE_AP_LOWER6_PRICE",
            "CUMUL_PRE_AP_LOWER60_PRICE",
            "CUMUL_PRE_AP_LOWER5MIN_PRICE",
            "CUMUL_PRE_AP_LOWERREG_PRICE",
        ]
        self.assertSequenceEqual(list(price_data.columns), expected_columns)

class TestWarnings(unittest.TestCase):
    def test_no_parse_warning(self):
        with patch('sys.stderr', new=StringIO()) as fakeOutput:
            print('hello world')
                
            dynamic_data_compiler(
                start_time='2017/01/01 00:00:00', 
                end_time='2017/01/01 00:05:00', 
                table_name='DISPATCHPRICE', 
                raw_data_location=defaults.raw_data_cache
            )

            stderr = fakeOutput.getvalue().strip()
        self.assertNotIn("UserWarning: Could not infer format", stderr)

class TestCacheCompilerRaisesExpectedErrors(unittest.TestCase):
    def test_raise_error_for_incorrect_table_name(self):
        with self.assertRaises(Exception) as context:
            cache_compiler(
                "2019/01/01 00:00:00",
                "2019/02/01 00:00:00",
                "NOTATABLE",
                defaults.raw_data_cache,
                fformat="db",
            )
        self.assertTrue(
            "Table name provided is not a dynamic table." in str(context.exception)
        )

    def test_raise_error_if_fformat_not_in_expected_set(self):
        with self.assertRaises(Exception) as context:
            cache_compiler(
                "2019/01/01 00:00:00",
                "2019/02/01 00:00:00",
                "DISPATCHPRICE",
                defaults.raw_data_cache,
                fformat="db",
            )
        self.assertTrue(
            "Argument fformat must be 'feather' or 'parquet'" in str(context.exception)
        )

    def test_raise_error_if_select_columns_used_without_rebuild_true(self):
        with self.assertRaises(Exception) as context:
            cache_compiler(
                "2019/01/01 00:00:00",
                "2019/02/01 00:00:00",
                "DISPATCHPRICE",
                defaults.raw_data_cache,
                select_columns="all",
            )
        self.assertTrue(
            (
                "The select_columns argument must be used with rebuild=True "
                + "to ensure the cache is built with the correct columns."
            )
            in str(context.exception)
        )



class TestStaticTableRaisesExpectedErrors(unittest.TestCase):
    def test_raise_error_for_incorrect_table_name(self):
        with self.assertRaises(Exception) as context:
            static_table("NOTATABLE", defaults.raw_data_cache)
        self.assertTrue(
            "Table name provided is not a static table." in str(context.exception)
        )

    def test_raise_error_for_no_data_returned(self):
        good_url = defaults.static_table_url["VARIABLES_FCAS_4_SECOND"]
        defaults.static_table_url["VARIABLES_FCAS_4_SECOND"] = "bad_url"
        path_and_name = (
                defaults.raw_data_cache + "/" + defaults.names["VARIABLES_FCAS_4_SECOND"]
        )
        if os.path.isfile(path_and_name):
            os.remove(path_and_name)
        with self.assertRaises(Exception) as context:
            static_table("VARIABLES_FCAS_4_SECOND", defaults.raw_data_cache)
        self.assertTrue(
            (
                f"Compiling data for table VARIABLES_FCAS_4_SECOND failed. "
                + "This probably because none of the requested data "
                + "could be download from AEMO. Check your internet "
                + "connection and that the requested data is archived on: "
                + "https://nemweb.com.au see nemosis.defaults for table specific urls."
            )
            in str(context.exception)
        )
        defaults.static_table_url["VARIABLES_FCAS_4_SECOND"] = good_url

    def test_raise_error_for_filter_column_not_in_select_columns(self):
        with self.assertRaises(Exception) as context:
            static_table(
                "VARIABLES_FCAS_4_SECOND",
                defaults.raw_data_cache,
                select_columns=["VARIABLENUMBER"],
                filter_cols=["VARIABLETYPE"],
                filter_values=(["0"],),
            )
        self.assertTrue(
            (
                "Filter columns not valid. They must be a part of "
                + "select_columns or the table defaults."
            )
            in str(context.exception)
        )

    def test_raise_error_for_filter_column_not_in_default_columns(self):
        with self.assertRaises(Exception) as context:
            static_table(
                "VARIABLES_FCAS_4_SECOND",
                defaults.raw_data_cache,
                select_columns=["VARIABLENUMBER"],
                filter_cols=["NOTACOLUMN"],
                filter_values=(["0"],),
            )
        self.assertTrue(
            (
                "Filter columns not valid. They must be a part of "
                + "select_columns or the table defaults."
            )
            in str(context.exception)
        )

    def test_raise_error_if_select_columns_not_in_data(self):
        with self.assertRaises(Exception) as context:
            static_table(
                "VARIABLES_FCAS_4_SECOND",
                defaults.raw_data_cache,
                select_columns=["NOTACOLUMN"],
            )
        self.assertIn(
            (
                f"None of columns ['NOTACOLUMN'] are in D:/nemosis_test_cache\\Ancillary Services Market Causer Pays Variables File.csv. "
                "This may be caused by user input if the 'select_columns' "
                "argument is being used, or by changed AEMO data formats. "
                "This error can be avoided by using the argument select_columns='all'."
            ),
            str(context.exception)
        )

    def test_using_select_columns_all_does_not_raise_error(self):
        price_data = static_table(
            "VARIABLES_FCAS_4_SECOND", defaults.raw_data_cache, select_columns="all"
        )
        expected_columns = ["VARIABLENUMBER", "VARIABLETYPE"]
        self.assertSequenceEqual(list(price_data.columns), expected_columns)



import unittest
from datetime import timedelta
from src.nemosis import defaults, data_fetch_methods
import pandas as pd
import os


class TestFormatOptions(unittest.TestCase):
    def setUp(self):
        # TODO: Clean tests since only one table - BIDDAYOFFER_D is tested
        self.table_names = ["BIDDAYOFFER_D"]

        self.table_types = {
            "DISPATCHLOAD": "DUID",
            "DISPATCHCONSTRAINT": "CONSTRAINTID",
            "DISPATCH_UNIT_SCADA": "DUID",
            "DISPATCHPRICE": "REGIONID",
            "DISPATCHINTERCONNECTORRES": "INTERCONNECTORID",
            "DISPATCHREGIONSUM": "REGIONID",
            "BIDPEROFFER_D": "DUID-BIDTYPE",
            "BIDDAYOFFER_D": "DUID-BIDTYPE",
            "TRADINGLOAD": "DUID",
            "TRADINGPRICE": "REGIONID",
            "TRADINGREGIONSUM": "REGIONID",
            "TRADINGINTERCONNECT": "INTERCONNECTORID",
        }

        self.filter_values = {
            "DUID": (["AGLHAL"],),
            "REGIONID": (["SA1"],),
            "INTERCONNECTORID": (["VIC1-NSW1"],),
            "CONSTRAINTID": (["DATASNAP_DFS_Q_CLST"],),
            "DUID-BIDTYPE": (["AGLHAL", "ENERGY"],),
        }

    def test_dispatch_tables_start_of_month_just_csv_format_dont_keep(self):
        # Empty cache.
        for f in os.listdir(defaults.raw_data_cache):
            os.remove(os.path.join(defaults.raw_data_cache, f))

        start_time = "2018/02/01 00:00:00"
        end_time = "2018/02/01 05:15:00"
        for table in self.table_names:
            print("Testing {} returing values at start of month.".format(table))
            dat_col = defaults.primary_date_columns[table]
            table_type = self.table_types[table]
            cols = [dat_col, self.table_types[table]]
            filter_cols = (self.table_types[table],)
            expected_length = 63
            expected_number_of_columns = 2
            expected_first_time = pd.to_datetime(
                start_time, format="%Y/%m/%d %H:%M:%S"
            ) + timedelta(minutes=5)
            expected_last_time = pd.to_datetime(end_time, format="%Y/%m/%d %H:%M:%S")
            if table in [
                "TRADINGLOAD",
                "TRADINGPRICE",
                "TRADINGREGIONSUM",
                "TRADINGINTERCONNECT",
            ]:
                expected_length = 10
                expected_first_time = "2018/02/01 00:30:00"
                expected_first_time = pd.to_datetime(
                    expected_first_time, format="%Y/%m/%d %H:%M:%S"
                )
                expected_last_time = "2018/02/01 05:00:00"
                expected_last_time = pd.to_datetime(
                    expected_last_time, format="%Y/%m/%d %H:%M:%S"
                )
            if table == "BIDPEROFFER_D":
                cols = [dat_col, "DUID", "BIDTYPE"]
                filter_cols = ("DUID", "BIDTYPE")
                expected_number_of_columns = 3
            if table == "BIDDAYOFFER_D":
                cols = [dat_col, "DUID", "BIDTYPE"]
                filter_cols = ("DUID", "BIDTYPE")
                expected_number_of_columns = 3
                expected_length = 2
                expected_last_time = "2018/02/01 00:00:00"
                expected_last_time = pd.to_datetime(
                    expected_last_time, format="%Y/%m/%d %H:%M:%S"
                )
                expected_first_time = "2018/01/31 00:00:00"
                expected_first_time = pd.to_datetime(
                    expected_first_time, format="%Y/%m/%d %H:%M:%S"
                )
            data = data_fetch_methods.dynamic_data_compiler(
                start_time,
                end_time,
                table,
                defaults.raw_data_cache,
                select_columns=cols,
                filter_cols=filter_cols,
                filter_values=self.filter_values[table_type],
                fformat="csv",
                keep_csv=False,
            )
            data = data.reset_index(drop=True)
            print(table)
            self.assertEqual(expected_length, data.shape[0])
            self.assertEqual(expected_number_of_columns, data.shape[1])
            self.assertEqual(expected_first_time, data[dat_col][0])
            self.assertEqual(expected_last_time, data[dat_col].iloc[-1])
            self.assertEqual(len(os.listdir(defaults.raw_data_cache)), 0)
            print("Passed")

    def test_dispatch_tables_start_of_month_just_csv_format(self):
        # Empty cache.
        for f in os.listdir(defaults.raw_data_cache):
            os.remove(os.path.join(defaults.raw_data_cache, f))

        start_time = "2018/02/01 00:00:00"
        end_time = "2018/02/01 05:15:00"
        for table in self.table_names:
            print("Testing {} returing values at start of month.".format(table))
            dat_col = defaults.primary_date_columns[table]
            table_type = self.table_types[table]
            cols = [dat_col, self.table_types[table]]
            filter_cols = (self.table_types[table],)
            expected_length = 63
            expected_number_of_columns = 2
            expected_first_time = pd.to_datetime(
                start_time, format="%Y/%m/%d %H:%M:%S"
            ) + timedelta(minutes=5)
            expected_last_time = pd.to_datetime(end_time, format="%Y/%m/%d %H:%M:%S")
            if table in [
                "TRADINGLOAD",
                "TRADINGPRICE",
                "TRADINGREGIONSUM",
                "TRADINGINTERCONNECT",
            ]:
                expected_length = 10
                expected_first_time = "2018/02/01 00:30:00"
                expected_first_time = pd.to_datetime(
                    expected_first_time, format="%Y/%m/%d %H:%M:%S"
                )
                expected_last_time = "2018/02/01 05:00:00"
                expected_last_time = pd.to_datetime(
                    expected_last_time, format="%Y/%m/%d %H:%M:%S"
                )
            if table == "BIDPEROFFER_D":
                cols = [dat_col, "DUID", "BIDTYPE"]
                filter_cols = ("DUID", "BIDTYPE")
                expected_number_of_columns = 3
            if table == "BIDDAYOFFER_D":
                cols = [dat_col, "DUID", "BIDTYPE"]
                filter_cols = ("DUID", "BIDTYPE")
                expected_number_of_columns = 3
                expected_length = 2
                expected_last_time = "2018/02/01 00:00:00"
                expected_last_time = pd.to_datetime(
                    expected_last_time, format="%Y/%m/%d %H:%M:%S"
                )
                expected_first_time = "2018/01/31 00:00:00"
                expected_first_time = pd.to_datetime(
                    expected_first_time, format="%Y/%m/%d %H:%M:%S"
                )
            data = data_fetch_methods.dynamic_data_compiler(
                start_time,
                end_time,
                table,
                defaults.raw_data_cache,
                select_columns=cols,
                filter_cols=filter_cols,
                filter_values=self.filter_values[table_type],
                fformat="csv",
            )
            data = data.reset_index(drop=True)
            print(table)
            self.assertEqual(expected_length, data.shape[0])
            self.assertEqual(expected_number_of_columns, data.shape[1])
            self.assertEqual(expected_first_time, data[dat_col][0])
            self.assertEqual(expected_last_time, data[dat_col].iloc[-1])
            self.assertNotEqual(len(os.listdir(defaults.raw_data_cache)), 0)
            print("Passed")

            # Test that also works on second pass.
            start_time = "2018/02/01 00:00:00"
            end_time = "2018/02/01 05:15:00"
            for table in self.table_names:
                print("Testing {} returing values at start of month.".format(table))
                dat_col = defaults.primary_date_columns[table]
                table_type = self.table_types[table]
                cols = [dat_col, self.table_types[table]]
                filter_cols = (self.table_types[table],)
                expected_length = 63
                expected_number_of_columns = 2
                expected_first_time = pd.to_datetime(
                    start_time, format="%Y/%m/%d %H:%M:%S"
                ) + timedelta(minutes=5)
                expected_last_time = pd.to_datetime(
                    end_time, format="%Y/%m/%d %H:%M:%S"
                )
                if table in [
                    "TRADINGLOAD",
                    "TRADINGPRICE",
                    "TRADINGREGIONSUM",
                    "TRADINGINTERCONNECT",
                ]:
                    expected_length = 10
                    expected_first_time = "2018/02/01 00:30:00"
                    expected_first_time = pd.to_datetime(
                        expected_first_time, format="%Y/%m/%d %H:%M:%S"
                    )
                    expected_last_time = "2018/02/01 05:00:00"
                    expected_last_time = pd.to_datetime(
                        expected_last_time, format="%Y/%m/%d %H:%M:%S"
                    )
                if table == "BIDPEROFFER_D":
                    cols = [dat_col, "DUID", "BIDTYPE"]
                    filter_cols = ("DUID", "BIDTYPE")
                    expected_number_of_columns = 3
                if table == "BIDDAYOFFER_D":
                    cols = [dat_col, "DUID", "BIDTYPE"]
                    filter_cols = ("DUID", "BIDTYPE")
                    expected_number_of_columns = 3
                    expected_length = 2
                    expected_last_time = "2018/02/01 00:00:00"
                    expected_last_time = pd.to_datetime(
                        expected_last_time, format="%Y/%m/%d %H:%M:%S"
                    )
                    expected_first_time = "2018/01/31 00:00:00"
                    expected_first_time = pd.to_datetime(
                        expected_first_time, format="%Y/%m/%d %H:%M:%S"
                    )
                data = data_fetch_methods.dynamic_data_compiler(
                    start_time,
                    end_time,
                    table,
                    defaults.raw_data_cache,
                    select_columns=cols,
                    filter_cols=filter_cols,
                    filter_values=self.filter_values[table_type],
                    fformat="csv",
                )
                data = data.reset_index(drop=True)
                print(table)
                self.assertEqual(expected_length, data.shape[0])
                self.assertEqual(expected_number_of_columns, data.shape[1])
                self.assertEqual(expected_first_time, data[dat_col][0])
                self.assertEqual(expected_last_time, data[dat_col].iloc[-1])
                self.assertNotEqual(len(os.listdir(defaults.raw_data_cache)), 0)
                print("Passed")

    def test_dispatch_tables_start_of_month_feather_format(self):
        start_time = "2018/02/01 00:00:00"
        end_time = "2018/02/01 05:15:00"
        for table in self.table_names:
            print("Testing {} returing values at start of month.".format(table))
            dat_col = defaults.primary_date_columns[table]
            table_type = self.table_types[table]
            cols = [dat_col, self.table_types[table]]
            filter_cols = (self.table_types[table],)
            expected_length = 63
            expected_number_of_columns = 2
            expected_first_time = pd.to_datetime(
                start_time, format="%Y/%m/%d %H:%M:%S"
            ) + timedelta(minutes=5)
            expected_last_time = pd.to_datetime(end_time, format="%Y/%m/%d %H:%M:%S")
            if table in [
                "TRADINGLOAD",
                "TRADINGPRICE",
                "TRADINGREGIONSUM",
                "TRADINGINTERCONNECT",
            ]:
                expected_length = 10
                expected_first_time = "2018/02/01 00:30:00"
                expected_first_time = pd.to_datetime(
                    expected_first_time, format="%Y/%m/%d %H:%M:%S"
                )
                expected_last_time = "2018/02/01 05:00:00"
                expected_last_time = pd.to_datetime(
                    expected_last_time, format="%Y/%m/%d %H:%M:%S"
                )
            if table == "BIDPEROFFER_D":
                cols = [dat_col, "DUID", "BIDTYPE"]
                filter_cols = ("DUID", "BIDTYPE")
                expected_number_of_columns = 3
            if table == "BIDDAYOFFER_D":
                cols = [dat_col, "DUID", "BIDTYPE"]
                filter_cols = ("DUID", "BIDTYPE")
                expected_number_of_columns = 3
                expected_length = 2
                expected_last_time = "2018/02/01 00:00:00"
                expected_last_time = pd.to_datetime(
                    expected_last_time, format="%Y/%m/%d %H:%M:%S"
                )
                expected_first_time = "2018/01/31 00:00:00"
                expected_first_time = pd.to_datetime(
                    expected_first_time, format="%Y/%m/%d %H:%M:%S"
                )
            print(table)
            data = data_fetch_methods.dynamic_data_compiler(
                start_time,
                end_time,
                table,
                defaults.raw_data_cache,
                select_columns=cols,
                filter_cols=filter_cols,
                filter_values=self.filter_values[table_type],
                fformat="feather",
            )
            data = data.reset_index(drop=True)
            print(table)
            self.assertEqual(expected_length, data.shape[0])
            self.assertEqual(expected_number_of_columns, data.shape[1])
            self.assertEqual(expected_first_time, data[dat_col][0])
            self.assertEqual(expected_last_time, data[dat_col].iloc[-1])
            print("Passed")

    def test_dispatch_tables_start_of_month_parquet_format(self):
        start_time = "2018/02/01 00:00:00"
        end_time = "2018/02/01 05:15:00"
        for table in self.table_names:
            print("Testing {} returing values at start of month.".format(table))
            dat_col = defaults.primary_date_columns[table]
            table_type = self.table_types[table]
            cols = [dat_col, self.table_types[table]]
            filter_cols = (self.table_types[table],)
            expected_length = 63
            expected_number_of_columns = 2
            expected_first_time = pd.to_datetime(
                start_time, format="%Y/%m/%d %H:%M:%S"
            ) + timedelta(minutes=5)
            expected_last_time = pd.to_datetime(end_time, format="%Y/%m/%d %H:%M:%S")
            if table in [
                "TRADINGLOAD",
                "TRADINGPRICE",
                "TRADINGREGIONSUM",
                "TRADINGINTERCONNECT",
            ]:
                expected_length = 10
                expected_first_time = "2018/02/01 00:30:00"
                expected_first_time = pd.to_datetime(
                    expected_first_time, format="%Y/%m/%d %H:%M:%S"
                )
                expected_last_time = "2018/02/01 05:00:00"
                expected_last_time = pd.to_datetime(
                    expected_last_time, format="%Y/%m/%d %H:%M:%S"
                )
            if table == "BIDPEROFFER_D":
                cols = [dat_col, "DUID", "BIDTYPE"]
                filter_cols = ("DUID", "BIDTYPE")
                expected_number_of_columns = 3
            if table == "BIDDAYOFFER_D":
                cols = [dat_col, "DUID", "BIDTYPE"]
                filter_cols = ("DUID", "BIDTYPE")
                expected_number_of_columns = 3
                expected_length = 2
                expected_last_time = "2018/02/01 00:00:00"
                expected_last_time = pd.to_datetime(
                    expected_last_time, format="%Y/%m/%d %H:%M:%S"
                )
                expected_first_time = "2018/01/31 00:00:00"
                expected_first_time = pd.to_datetime(
                    expected_first_time, format="%Y/%m/%d %H:%M:%S"
                )
            data = data_fetch_methods.dynamic_data_compiler(
                start_time,
                end_time,
                table,
                defaults.raw_data_cache,
                select_columns=cols,
                filter_cols=filter_cols,
                filter_values=self.filter_values[table_type],
                fformat="parquet",
                parse_data_types=True,
            )
            data = data.reset_index(drop=True)
            print(table)
            self.assertEqual(expected_length, data.shape[0])
            self.assertEqual(expected_number_of_columns, data.shape[1])
            self.assertEqual(expected_first_time, data[dat_col][0])
            self.assertEqual(expected_last_time, data[dat_col].iloc[-1])
            self.assertFalse(all(object == data.dtypes))
            print("Passed")



import tkinter as tk
import tkinter.ttk as ttk
from nemosis import defaults


class Query:
    def __init__(self, master, row_number, app, table_options=None):
        # Load in the starting features of a query row.
        self.master = master
        self.row_number = row_number

        # Make an label and entry box for the user to name the query result.
        self.query_label = ttk.Label(self.master, text="  Query name:")
        self.name = ttk.Entry(self.master)
        self.name.config(width=26)

        # Make labels and entry boxes for the user to provide start and end time to filter the query based on.
        self.start_time_label = ttk.Label(
            self.master, text="Start time:\n(YYYY/MM/DD HH:MM:SS)"
        )
        self.start_time = ttk.Entry(self.master)
        self.start_time.config(width=26)
        self.end_time_label = ttk.Label(
            self.master, text="End time:\n(YYYY/MM/DD HH:MM:SS)"
        )
        self.end_time = ttk.Entry(self.master)
        self.end_time.config(width=26)

        # Create a label and a list of tables to choose from.
        self.tables_label = ttk.Label(self.master, text="Select table:")
        self.tables = tk.Listbox(self.master, exportselection=False, width=35)
        self.tables.bind("<<ListboxSelect>>", self.add_column_selection)

        self.table_options = table_options

        for item in table_options:
            self.tables.insert(tk.END, item)

        # Create a button to delete the row.
        self.delete = ttk.Button(
            self.master, text="\u274C", command=lambda: app.delete_row(self.row_number)
        )

        # Create empty attributes to fill up later on.
        self.filter_list = {}
        self.filter_label = {}
        self.filter_entry = {}
        self.col_list = None
        self.cols_label = None

        # Position all the widgets in the row.
        self.position()

    def position(self):
        first_sub_column = 0
        second_sub_column = 1
        pady = defaults.query_y_pad
        padx = defaults.standard_x_pad

        self.query_label.grid(
            row=defaults.query_row_offset + defaults.row_height * self.row_number,
            column=first_sub_column,
            pady=pady,
            padx=padx,
            sticky="sw",
        )
        self.query_label.update()

        self.name.grid(
            row=defaults.query_row_offset
                + defaults.row_height * self.row_number
                + defaults.names_internal_row,
            column=first_sub_column,
            padx=padx,
            sticky="sw",
        )
        self.name.update()

        self.start_time_label.grid(
            row=defaults.query_row_offset
                + defaults.row_height * self.row_number
                + defaults.start_time_label_internal_row,
            column=first_sub_column,
            padx=defaults.standard_x_pad,
            sticky="sw",
        )
        self.start_time_label.update()

        self.start_time.grid(
            row=defaults.query_row_offset
                + defaults.row_height * self.row_number
                + defaults.start_time_internal_row,
            column=first_sub_column,
            padx=padx,
            sticky="sw",
        )
        self.start_time.update()

        self.end_time_label.grid(
            row=defaults.query_row_offset
                + defaults.row_height * self.row_number
                + defaults.end_time_label_internal_row,
            column=first_sub_column,
            padx=padx,
            sticky="sw",
        )
        self.end_time_label.update()

        self.end_time.grid(
            row=defaults.query_row_offset
                + defaults.row_height * self.row_number
                + defaults.end_time_internal_row,
            column=first_sub_column,
            padx=padx,
            sticky="sw",
        )
        self.end_time.update()

        self.tables_label.grid(
            row=defaults.query_row_offset + defaults.row_height * self.row_number,
            column=second_sub_column,
            pady=defaults.query_y_pad,
            sticky="sw",
            padx=padx,
        )
        self.tables_label.update()

        self.tables.grid(
            row=defaults.query_row_offset
                + defaults.row_height * self.row_number
                + defaults.table_list_internal_row,
            column=second_sub_column,
            rowspan=defaults.list_row_span,
            columnspan=defaults.list_column_span,
            sticky="sw",
            padx=padx,
        )
        self.tables.update()

        self.delete.grid(
            row=defaults.query_row_offset
                + defaults.row_height * self.row_number
                + defaults.delete_button_internal_row,
            column=defaults.last_column,
            sticky="nw",
        )
        self.delete.update()

        if self.col_list is not None:
            self.position_column_list()
        if self.filter_list:
            self.position_filter_list()

    def add_column_selection(self, evt):
        # When a table is selected update the list of columns to be selected from.
        # Find the name of the table selected.
        table = self.table_options[self.tables.curselection()[0]]
        # Delete the previous list of columns.
        if self.col_list is not None:
            self.col_list.destroy()
            self.cols_label.destroy()

        # Create a new label and list box.
        self.cols_label = ttk.Label(self.master, text="Select columns:")
        self.col_list = tk.Listbox(
            self.master, selectmode=tk.MULTIPLE, exportselection=False, width=26
        )
        self.col_list.bind("<<ListboxSelect>>", self.add_filters)
        self.col_list.delete(0, tk.END)

        # Populate the list box with column names.
        for item in defaults.table_columns[table]:
            self.col_list.insert(tk.END, item)

        for col in defaults.table_primary_keys[table]:
            index = defaults.table_columns[table].index(col)
            self.col_list.selection_set(index)

        # Position the column list.
        self.position_column_list()

        # Delete any filters that existed for previous column selections.
        self.remove_filters()
        self.add_filters(None)

    def position_column_list(self):
        self.cols_label.grid(
            column=self.tables.grid_info()["column"] + defaults.list_column_span,
            row=defaults.query_row_offset + self.row_number * defaults.row_height,
            sticky="sw",
        )
        self.cols_label.update()
        self.col_list.grid(
            column=self.tables.grid_info()["column"] + defaults.list_column_span,
            row=defaults.query_row_offset
                + self.row_number * defaults.row_height
                + defaults.table_list_internal_row,
            rowspan=defaults.list_row_span,
            columnspan=defaults.list_column_span,
            padx=defaults.standard_x_pad,
            sticky="sw",
        )
        self.col_list.update()

    def add_filters(self, evt):
        # When a column is selected update the filter options.

        # Remove filters for columns that are no longer selected.
        self.remove_filters_unselected()

        # Find which columns are currently selected.
        select_cols = [
            self.col_list.get(0, tk.END)[index]
            for index in self.col_list.curselection()
        ]

        # If a column is selected, and is filterable, but does not have a filter then add a filter for that column.
        for column in select_cols:
            if (
                column in defaults.filterable_cols
                and column not in self.filter_label.keys()
            ):
                self.filter_label[column] = ttk.Label(
                    self.master, text="Select {}s:".format(str(column))
                )
                self.filter_entry[column] = ttk.Entry(self.master, width=25)
                self.filter_entry[column].bind("<Return>", self.add_to_list)
                self.filter_entry[column].name = column
                self.filter_list[column] = tk.Listbox(
                    self.master,
                    selectmode=tk.MULTIPLE,
                    exportselection=False,
                    height=8,
                    width=25,
                )

        # Position all the filters so there are no gaps in between them.
        self.position_filter_list()

    def position_filter_list(self):
        last_filter = None
        # Position all the filter next to each other moving left across the screen.
        for column in self.filter_label.keys():
            if last_filter is None:
                # Place the first filter next to the column list.
                col = self.col_list.grid_info()["column"] + defaults.list_column_span
            else:
                # Place the next filter next to the last filter.
                col = (
                        self.filter_label[last_filter].grid_info()["column"]
                        + defaults.list_column_span
                )

            self.filter_label[column].grid(
                row=defaults.query_row_offset + defaults.row_height * self.row_number,
                column=col,
                padx=defaults.standard_x_pad,
                sticky="sw",
            )
            self.filter_label[column].update()
            self.filter_entry[column].grid(
                row=defaults.query_row_offset
                    + defaults.row_height * self.row_number
                    + defaults.names_internal_row,
                column=col,
                padx=defaults.standard_x_pad,
            )
            self.filter_entry[column].update()
            self.filter_list[column].grid(
                row=defaults.query_row_offset
                    + defaults.row_height * self.row_number
                    + defaults.internal_filter_row,
                column=col,
                columnspan=defaults.list_column_span,
                rowspan=defaults.list_filter_row_span,
                padx=defaults.standard_x_pad,
            )
            self.filter_list[column].update()
            last_filter = column

    def add_to_list(self, evt):
        # Add the item in the entry box of a filter to the list box below.
        self.filter_list[evt.widget.name].insert(tk.END, evt.widget.get())
        evt.widget.delete(0, "end")

    def remove_filters_unselected(self):
        # Delete filter whoes columns are not selected.
        select_cols = [
            self.col_list.get(0, tk.END)[index]
            for index in self.col_list.curselection()
        ]
        existing_filters = list(self.filter_label.keys())
        for column in existing_filters:
            if column not in select_cols:
                self.filter_label[column].destroy()
                del self.filter_label[column]
                self.filter_entry[column].destroy()
                del self.filter_entry[column]
                self.filter_list[column].destroy()
                del self.filter_list[column]

    def empty(self):
        # Delete all the widgets in the row.
        if self.filter_list:
            # Delete the filters.
            self.remove_filters()
        if self.col_list is not None:
            # Delete the column list.
            self.col_list.destroy()
            del self.col_list
        if self.cols_label is not None:
            self.cols_label.destroy()
            del self.cols_label
        # Delete the the row features that always exist.
        self.remove_initial_features()

    def remove_initial_features(self):
        # Remove the initial widegts.
        self.query_label.destroy()
        del self.query_label
        self.name.destroy()
        del self.name
        self.start_time_label.destroy()
        del self.start_time_label
        self.start_time.destroy()
        del self.start_time
        self.end_time_label.destroy()
        del self.end_time_label
        self.end_time.destroy()
        del self.end_time
        self.tables_label.destroy()
        del self.tables_label
        self.tables.destroy()
        del self.tables
        self.delete.destroy()
        del self.delete

    def remove_filters(self):
        # Remove any filter widgets that exist.
        existing_filters = list(self.filter_label.keys())
        for column in existing_filters:
            self.filter_label[column].destroy()
            del self.filter_label[column]
            self.filter_entry[column].destroy()
            del self.filter_entry[column]
            self.filter_list[column].destroy()
            del self.filter_list[column]

    def state(self):
        # Return the current state of the row as a dictionary.
        state = {}
        state["type"] = "query"
        state["name"] = self.name.get()
        state["start_time"] = self.start_time.get()
        state["end_time"] = self.end_time.get()
        if len(self.tables.curselection()) != 0:
            state["table"] = self.table_options[self.tables.curselection()[0]]
        if self.col_list is not None:
            state["columns"] = [
                self.col_list.get(0, tk.END)[index]
                for index in self.col_list.curselection()
            ]
        state["filters_contents"] = {}
        state["filters_selection"] = {}
        for column, filter_list in self.filter_list.items():
            state["filters_contents"][column] = self.filter_list[column].get(0, tk.END)
            state["filters_selection"][column] = self.filter_list[column].curselection()
        return state

    def load_state(self, state):
        # Update the row to match the state provided.
        self.name.insert(0, state["name"])
        self.start_time.insert(0, state["start_time"])
        self.end_time.insert(0, state["end_time"])
        if len(state["table"]) != 0:
            table_index = list(self.tables.get(0, "end")).index(state["table"])
            self.tables.selection_set(table_index)
            self.add_column_selection(None)
            for col in state["columns"]:
                col_index = list(self.col_list.get(0, "end")).index(col)
                self.col_list.selection_set(col_index)
            self.add_filters(None)
        for column, filter_contents in state["filters_contents"].items():
            self.filter_list[column].insert(0, *filter_contents)
        for column, filter_selection in state["filters_selection"].items():
            for index in filter_selection:
                self.filter_list[column].selection_set(index)


class Merge_as_of:
    def __init__(self, master, row_number, app):
        # Create all the widgets of a merge row.
        self.master = master
        self.row_number = row_number
        # Create a label and entry box to name the result of the merge
        self.merge_label = ttk.Label(self.master, text="Merge name")
        self.name = ttk.Entry(self.master)
        self.name.config(width=26)
        # Create entry box to provide the name of the left result to merge.
        self.left_table_label = ttk.Label(self.master, text="Left table")
        self.left_table = ttk.Entry(self.master)
        self.left_table.config(width=26)
        # Create an entry box to provide the name of the right result to merge.
        self.right_table_label = ttk.Label(self.master, text="Right table")
        self.right_table = ttk.Entry(self.master)
        self.right_table.config(width=26)
        # Create a list to select the merge type from.
        self.join_types_label = ttk.Label(self.master, text="Select join type")
        self.join_types = tk.Listbox(self.master, exportselection=False, width=28)
        for item in defaults.join_type:
            self.join_types.insert(tk.END, item)
        # Create a button that deletes the row.
        self.delete = ttk.Button(
            self.master, text="\u274C", command=lambda: app.delete_row(self.row_number)
        )
        # Create a entry box and list to provide the keys to the left result.
        self.left_time_keys_label = ttk.Label(self.master, text="Left time keys")
        self.left_time_keys_entry = ttk.Entry(self.master)
        self.left_time_keys_entry.bind("<Return>", self.add_to_list_left_time)
        self.left_time_key_list = tk.Listbox(
            self.master, selectmode=tk.MULTIPLE, exportselection=False, height=8
        )
        # Create a entry box and list to provide the keys to the right result.
        self.right_time_keys_label = ttk.Label(self.master, text="Right time keys")
        self.right_time_keys_entry = ttk.Entry(self.master)
        self.right_time_keys_entry.bind("<Return>", self.add_to_list_right_time)
        self.right_time_key_list = tk.Listbox(
            self.master, selectmode=tk.MULTIPLE, exportselection=False, height=8
        )
        # Create a entry box and list to provide the keys to the left result.
        self.left_keys_label = ttk.Label(self.master, text="Left keys")
        self.left_keys_entry = ttk.Entry(self.master)
        self.left_keys_entry.bind("<Return>", self.add_to_list_left)
        self.left_key_list = tk.Listbox(
            self.master, selectmode=tk.MULTIPLE, exportselection=False, height=8
        )
        # Create a entry box and list to provide the keys to the right result.
        self.right_keys_label = ttk.Label(self.master, text="Right keys")
        self.right_keys_entry = ttk.Entry(self.master)
        self.right_keys_entry.bind("<Return>", self.add_to_list_right)
        self.right_key_list = tk.Listbox(
            self.master, selectmode=tk.MULTIPLE, exportselection=False, height=8
        )
        # Position all the widgets.
        self.position()

    def position(self):
        self.merge_label.grid(
            row=defaults.query_row_offset + defaults.row_height * self.row_number,
            column=0,
            pady=defaults.query_y_pad,
            padx=defaults.standard_x_pad,
            sticky="ws",
        )
        self.merge_label.update()
        self.name.grid(
            row=defaults.query_row_offset
                + defaults.row_height * self.row_number
                + defaults.names_internal_row,
            column=0,
            padx=defaults.standard_x_pad,
        )
        self.name.update()
        self.left_table_label.grid(
            row=defaults.query_row_offset
                + defaults.row_height * self.row_number
                + defaults.start_time_label_internal_row,
            column=0,
            padx=defaults.standard_x_pad,
        )
        self.left_table_label.update()
        self.left_table.grid(
            row=defaults.query_row_offset
                + defaults.row_height * self.row_number
                + defaults.start_time_internal_row,
            column=0,
            padx=defaults.standard_x_pad,
        )
        self.left_table.update()
        self.right_table_label.grid(
            row=defaults.query_row_offset
                + defaults.row_height * self.row_number
                + defaults.end_time_label_internal_row,
            column=0,
            padx=defaults.standard_x_pad,
        )
        self.right_table_label.update()
        self.right_table.grid(
            row=defaults.query_row_offset
                + defaults.row_height * self.row_number
                + defaults.end_time_internal_row,
            column=0,
            padx=defaults.standard_x_pad,
        )
        self.right_table.update()
        self.join_types_label.grid(
            row=defaults.query_row_offset + defaults.row_height * self.row_number,
            column=1,
            pady=defaults.query_y_pad,
            sticky="sw",
            padx=defaults.standard_x_pad,
        )
        self.join_types_label.update()
        self.join_types.grid(
            row=defaults.query_row_offset
                + defaults.row_height * self.row_number
                + defaults.table_list_internal_row,
            column=1,
            rowspan=defaults.list_row_span,
            columnspan=defaults.list_column_span,
            sticky="nw",
            padx=defaults.standard_x_pad,
        )
        self.join_types.update()
        self.delete.grid(
            row=defaults.query_row_offset
                + defaults.row_height * self.row_number
                + defaults.delete_button_internal_row,
            column=defaults.last_column,
            sticky="nw",
        )
        self.delete.update()

        label_row = defaults.query_row_offset + defaults.row_height * self.row_number
        label_sticky = "w"
        entry_row = (
                defaults.query_row_offset
                + defaults.row_height * self.row_number
                + defaults.names_internal_row
        )
        custom_list_row = (
                defaults.query_row_offset
                + defaults.row_height * self.row_number
                + defaults.internal_filter_row
        )

        left_time_keys_col = (
                self.join_types.grid_info()["column"] + defaults.list_column_span
        )
        self.left_time_keys_label.grid(
            row=label_row,
            column=left_time_keys_col,
            sticky=label_sticky,
            padx=defaults.standard_x_pad,
        )
        self.left_time_keys_entry.grid(
            row=entry_row,
            column=left_time_keys_col,
            rowspan=defaults.list_row_span,
            columnspan=defaults.list_column_span,
            sticky="nw",
            padx=defaults.standard_x_pad,
        )
        self.left_time_key_list.grid(
            row=custom_list_row,
            column=left_time_keys_col,
            rowspan=defaults.list_row_span,
            columnspan=defaults.list_column_span,
            sticky="nw",
            padx=defaults.standard_x_pad,
        )

        right_time_keys_col = (
                self.left_time_key_list.grid_info()["column"] + defaults.list_column_span
        )
        self.right_time_keys_label.grid(
            row=label_row,
            column=right_time_keys_col,
            sticky=label_sticky,
            padx=defaults.standard_x_pad,
        )
        self.right_time_keys_entry.grid(
            row=entry_row,
            column=right_time_keys_col,
            rowspan=defaults.list_row_span,
            columnspan=defaults.list_column_span,
            sticky="nw",
            padx=defaults.standard_x_pad,
        )
        self.right_time_key_list.grid(
            row=custom_list_row,
            column=right_time_keys_col,
            rowspan=defaults.list_row_span,
            columnspan=defaults.list_column_span,
            sticky="nw",
            padx=defaults.standard_x_pad,
        )

        left_keys_col = (
                self.right_time_key_list.grid_info()["column"] + defaults.list_column_span
        )
        self.left_keys_label.grid(
            row=label_row,
            column=left_keys_col,
            sticky=label_sticky,
            padx=defaults.standard_x_pad,
        )
        self.left_keys_entry.grid(
            row=entry_row,
            column=left_keys_col,
            rowspan=defaults.list_row_span,
            columnspan=defaults.list_column_span,
            sticky="nw",
            padx=defaults.standard_x_pad,
        )
        self.left_key_list.grid(
            row=custom_list_row,
            column=left_keys_col,
            rowspan=defaults.list_row_span,
            columnspan=defaults.list_column_span,
            sticky="nw",
            padx=defaults.standard_x_pad,
        )

        right_keys_col = (
                self.left_key_list.grid_info()["column"] + defaults.list_column_span
        )
        self.right_keys_label.grid(
            row=label_row,
            column=right_keys_col,
            sticky=label_sticky,
            padx=defaults.standard_x_pad,
        )
        self.right_keys_entry.grid(
            row=entry_row,
            column=right_keys_col,
            rowspan=defaults.list_row_span,
            columnspan=defaults.list_column_span,
            sticky="nw",
            padx=defaults.standard_x_pad,
        )
        self.right_key_list.grid(
            row=custom_list_row,
            column=right_keys_col,
            rowspan=defaults.list_row_span,
            columnspan=defaults.list_column_span,
            sticky="nw",
            padx=defaults.standard_x_pad,
        )

    def add_to_list_left_time(self, evt):
        # Add key from entry box to list.
        self.left_time_key_list.insert(tk.END, evt.widget.get())
        evt.widget.delete(0, "end")

    def add_to_list_right_time(self, evt):
        # Add key from entry box to list.
        self.right_time_key_list.insert(tk.END, evt.widget.get())
        evt.widget.delete(0, "end")

    def add_to_list_left(self, evt):
        # Add key from entry box to list.
        self.left_key_list.insert(tk.END, evt.widget.get())
        evt.widget.delete(0, "end")

    def add_to_list_right(self, evt):
        # Add key from entry box to list.
        self.right_key_list.insert(tk.END, evt.widget.get())
        evt.widget.delete(0, "end")

    def empty(self):
        # Delete the widgets of the merge row.
        self.merge_label.destroy()
        del self.merge_label
        self.name.destroy()
        del self.name
        self.left_table_label.destroy()
        del self.left_table_label
        self.left_table.destroy()
        del self.left_table
        self.right_table_label.destroy()
        del self.right_table_label
        self.right_table.destroy()
        del self.right_table
        self.join_types_label.destroy()
        del self.join_types_label
        self.join_types.destroy()
        del self.join_types
        self.left_time_keys_label.destroy()
        del self.left_time_keys_label
        self.left_time_keys_entry.destroy()
        del self.left_time_keys_entry
        self.left_time_key_list.destroy()
        del self.left_time_key_list
        self.right_time_keys_label.destroy()
        del self.right_time_keys_label
        self.right_time_keys_entry.destroy()
        del self.right_time_keys_entry
        self.right_time_key_list.destroy()
        del self.right_time_key_list
        self.left_keys_label.destroy()
        del self.left_keys_label
        self.left_keys_entry.destroy()
        del self.left_keys_entry
        self.left_key_list.destroy()
        del self.left_key_list
        self.right_keys_label.destroy()
        del self.right_keys_label
        self.right_keys_entry.destroy()
        del self.right_keys_entry
        self.right_key_list.destroy()
        del self.right_key_list
        self.delete.destroy()
        del self.delete

    def state(self):
        # Return the state of the row as a dictionary.
        state = {}
        state["type"] = "merge_as_of"
        state["name"] = self.name.get()
        state["left_table"] = self.left_table.get()
        state["right_table"] = self.right_table.get()
        state["join_types"] = self.join_types.curselection()
        state["left_time_key_list"] = {}
        state["left_time_key_list"]["contents"] = self.left_time_key_list.get(0, tk.END)
        state["left_time_key_list"][
            "selection"
        ] = self.left_time_key_list.curselection()
        state["right_time_key_list"] = {}
        state["right_time_key_list"]["contents"] = self.right_time_key_list.get(
            0, tk.END
        )
        state["right_time_key_list"][
            "selection"
        ] = self.right_time_key_list.curselection()
        state["left_key_list"] = {}
        state["left_key_list"]["contents"] = self.left_key_list.get(0, tk.END)
        state["left_key_list"]["selection"] = self.left_key_list.curselection()
        state["right_key_list"] = {}
        state["right_key_list"]["contents"] = self.right_key_list.get(0, tk.END)
        state["right_key_list"]["selection"] = self.right_key_list.curselection()
        return state

    def load_state(self, state):
        # Update the row to match the state provided.
        self.name.insert(0, state["name"])
        self.left_table.insert(0, state["left_table"])
        self.right_table.insert(0, state["right_table"])
        if len(state["join_types"]) != 0:
            self.join_types.selection_set(state["join_types"][0])

        self.left_time_key_list.insert(0, *state["left_time_key_list"]["contents"])
        for index in state["left_time_key_list"]["selection"]:
            self.left_time_key_list.selection_set(index)

        self.right_time_key_list.insert(0, *state["right_time_key_list"]["contents"])
        for index in state["right_time_key_list"]["selection"]:
            self.right_time_key_list.selection_set(index)

        self.left_key_list.insert(0, *state["left_key_list"]["contents"])
        for index in state["left_key_list"]["selection"]:
            self.left_key_list.selection_set(index)

        self.right_key_list.insert(0, *state["right_key_list"]["contents"])
        for index in state["right_key_list"]["selection"]:
            self.right_key_list.selection_set(index)


class FilterVersionNo:
    def __init__(self, master, row_number, app):
        # Create all the widgets of a merge row.
        self.master = master
        self.row_number = row_number
        # Create a label and entry box to name the result of filter
        self.output_label = ttk.Label(self.master, text="Output table")
        self.name = ttk.Entry(self.master)
        self.name.config(width=26)
        # Create entry box to provide the name of the input.
        self.input_label = ttk.Label(self.master, text="Input table")
        self.input = ttk.Entry(self.master)
        self.input.config(width=26)
        # Create a button that deletes the row.
        self.delete = ttk.Button(
            self.master, text="\u274C", command=lambda: app.delete_row(self.row_number)
        )
        self.position()

    def position(self):
        self.output_label.grid(
            row=defaults.query_row_offset + defaults.row_height * self.row_number,
            column=0,
            pady=defaults.query_y_pad,
            padx=defaults.standard_x_pad,
            sticky="ws",
        )
        self.output_label.update()
        self.name.grid(
            row=defaults.query_row_offset
                + defaults.row_height * self.row_number
                + defaults.names_internal_row,
            column=0,
            padx=defaults.standard_x_pad,
        )
        self.name.update()

        self.input_label.grid(
            row=defaults.query_row_offset
                + defaults.row_height * self.row_number
                + defaults.start_time_label_internal_row,
            column=0,
            padx=defaults.standard_x_pad,
            sticky="ws",
        )
        self.input_label.update()
        self.input.grid(
            row=defaults.query_row_offset
                + defaults.row_height * self.row_number
                + defaults.start_time_internal_row,
            column=0,
            padx=defaults.standard_x_pad,
        )
        self.input.update()
        self.delete.grid(
            row=defaults.query_row_offset
                + defaults.row_height * self.row_number
                + defaults.delete_button_internal_row,
            column=defaults.last_column,
            sticky="nw",
        )
        self.delete.update()

    def empty(self):
        # Delete the widgets of the merge row.
        self.output_label.destroy()
        del self.output_label
        self.name.destroy()
        del self.name
        self.input_label.destroy()
        del self.input_label
        self.input.destroy()
        del self.input
        self.delete.destroy()
        del self.delete

    def state(self):
        # Return the state of the row as a dictionary.
        state = {}
        state["type"] = "filter_version_no"
        state["name"] = self.name.get()
        state["input"] = self.input.get()
        return state

    def load_state(self, state):
        # Update the row to match the state provided.
        self.name.insert(0, state["name"])
        self.input.insert(0, state["input"])


class Merge:
    def __init__(self, master, row_number, app):
        # Create all the widgets of a merge row.
        self.master = master
        self.row_number = row_number
        # Create a label and entry box to name the result of the merge
        self.merge_label = ttk.Label(self.master, text="Merge name")
        self.name = ttk.Entry(self.master)
        self.name.config(width=26)
        # Create entry box to provide the name of the left result to merge.
        self.left_table_label = ttk.Label(self.master, text="Left table")
        self.left_table = ttk.Entry(self.master)
        self.left_table.config(width=26)
        # Create an entry box to provide the name of the right result to merge.
        self.right_table_label = ttk.Label(self.master, text="Right table")
        self.right_table = ttk.Entry(self.master)
        self.right_table.config(width=26)
        # Create a list to select the merge type from.
        self.join_types_label = ttk.Label(self.master, text="Select join type")
        self.join_types = tk.Listbox(self.master, exportselection=False, width=28)
        for item in defaults.join_type:
            self.join_types.insert(tk.END, item)
        # Create a button that deletes the row.
        self.delete = ttk.Button(
            self.master, text="\u274C", command=lambda: app.delete_row(self.row_number)
        )
        # Create a entry box and list to provide the keys to the left result.
        self.left_keys_label = ttk.Label(self.master, text="Left keys")
        self.left_keys_entry = ttk.Entry(self.master)
        self.left_keys_entry.bind("<Return>", self.add_to_list_left)
        self.left_key_list = tk.Listbox(
            self.master, selectmode=tk.MULTIPLE, exportselection=False, height=8
        )
        # Create a entry box and list to provide the keys to the right result.
        self.right_keys_label = ttk.Label(self.master, text="Right keys")
        self.right_keys_entry = ttk.Entry(self.master)
        self.right_keys_entry.bind("<Return>", self.add_to_list_right)
        self.right_key_list = tk.Listbox(
            self.master, selectmode=tk.MULTIPLE, exportselection=False, height=8
        )
        # Position all the widgets.
        self.position()

    def position(self):
        self.merge_label.grid(
            row=defaults.query_row_offset + defaults.row_height * self.row_number,
            column=0,
            pady=defaults.query_y_pad,
            padx=defaults.standard_x_pad,
            sticky="ws",
        )
        self.merge_label.update()
        self.name.grid(
            row=defaults.query_row_offset
                + defaults.row_height * self.row_number
                + defaults.names_internal_row,
            column=0,
            padx=defaults.standard_x_pad,
        )
        self.name.update()
        self.left_table_label.grid(
            row=defaults.query_row_offset
                + defaults.row_height * self.row_number
                + defaults.start_time_label_internal_row,
            column=0,
            padx=defaults.standard_x_pad,
        )
        self.left_table_label.update()
        self.left_table.grid(
            row=defaults.query_row_offset
                + defaults.row_height * self.row_number
                + defaults.start_time_internal_row,
            column=0,
            padx=defaults.standard_x_pad,
        )
        self.left_table.update()
        self.right_table_label.grid(
            row=defaults.query_row_offset
                + defaults.row_height * self.row_number
                + defaults.end_time_label_internal_row,
            column=0,
            padx=defaults.standard_x_pad,
        )
        self.right_table_label.update()
        self.right_table.grid(
            row=defaults.query_row_offset
                + defaults.row_height * self.row_number
                + defaults.end_time_internal_row,
            column=0,
            padx=defaults.standard_x_pad,
        )
        self.right_table.update()
        self.join_types_label.grid(
            row=defaults.query_row_offset + defaults.row_height * self.row_number,
            column=1,
            pady=defaults.query_y_pad,
            sticky="sw",
            padx=defaults.standard_x_pad,
        )
        self.join_types_label.update()
        self.join_types.grid(
            row=defaults.query_row_offset
                + defaults.row_height * self.row_number
                + defaults.table_list_internal_row,
            column=1,
            rowspan=defaults.list_row_span,
            columnspan=defaults.list_column_span,
            sticky="nw",
            padx=defaults.standard_x_pad,
        )
        self.join_types.update()
        self.delete.grid(
            row=defaults.query_row_offset
                + defaults.row_height * self.row_number
                + defaults.delete_button_internal_row,
            column=defaults.last_column,
            sticky="nw",
        )
        self.delete.update()

        label_row = defaults.query_row_offset + defaults.row_height * self.row_number
        label_sticky = "w"
        entry_row = (
                defaults.query_row_offset
                + defaults.row_height * self.row_number
                + defaults.names_internal_row
        )
        custom_list_row = (
                defaults.query_row_offset
                + defaults.row_height * self.row_number
                + defaults.internal_filter_row
        )

        left_keys_col = (
                self.join_types.grid_info()["column"] + defaults.list_column_span
        )
        self.left_keys_label.grid(
            row=label_row,
            column=left_keys_col,
            sticky=label_sticky,
            padx=defaults.standard_x_pad,
        )
        self.left_keys_entry.grid(
            row=entry_row,
            column=left_keys_col,
            rowspan=defaults.list_row_span,
            columnspan=defaults.list_column_span,
            sticky="nw",
            padx=defaults.standard_x_pad,
        )
        self.left_key_list.grid(
            row=custom_list_row,
            column=left_keys_col,
            rowspan=defaults.list_row_span,
            columnspan=defaults.list_column_span,
            sticky="nw",
            padx=defaults.standard_x_pad,
        )

        right_keys_col = (
                self.left_key_list.grid_info()["column"] + defaults.list_column_span
        )
        self.right_keys_label.grid(
            row=label_row,
            column=right_keys_col,
            sticky=label_sticky,
            padx=defaults.standard_x_pad,
        )
        self.right_keys_entry.grid(
            row=entry_row,
            column=right_keys_col,
            rowspan=defaults.list_row_span,
            columnspan=defaults.list_column_span,
            sticky="nw",
            padx=defaults.standard_x_pad,
        )
        self.right_key_list.grid(
            row=custom_list_row,
            column=right_keys_col,
            rowspan=defaults.list_row_span,
            columnspan=defaults.list_column_span,
            sticky="nw",
            padx=defaults.standard_x_pad,
        )

    def add_to_list_left(self, evt):
        # Add key from entry box to list.
        self.left_key_list.insert(tk.END, evt.widget.get())
        evt.widget.delete(0, "end")

    def add_to_list_right(self, evt):
        # Add key from entry box to list.
        self.right_key_list.insert(tk.END, evt.widget.get())
        evt.widget.delete(0, "end")

    def empty(self):
        # Delete the widgets of the merge row.
        self.merge_label.destroy()
        del self.merge_label
        self.name.destroy()
        del self.name
        self.left_table_label.destroy()
        del self.left_table_label
        self.left_table.destroy()
        del self.left_table
        self.right_table_label.destroy()
        del self.right_table_label
        self.right_table.destroy()
        del self.right_table
        self.join_types_label.destroy()
        del self.join_types_label
        self.join_types.destroy()
        del self.join_types
        self.left_keys_label.destroy()
        del self.left_keys_label
        self.left_keys_entry.destroy()
        del self.left_keys_entry
        self.left_key_list.destroy()
        del self.left_key_list
        self.right_keys_label.destroy()
        del self.right_keys_label
        self.right_keys_entry.destroy()
        del self.right_keys_entry
        self.right_key_list.destroy()
        del self.right_key_list
        self.delete.destroy()
        del self.delete

    def state(self):
        # Return the state of the row as a dictionary.
        state = {}
        state["type"] = "merge"
        state["name"] = self.name.get()
        state["left_table"] = self.left_table.get()
        state["right_table"] = self.right_table.get()
        state["join_types"] = self.join_types.curselection()
        state["left_key_list"] = {}
        state["left_key_list"]["contents"] = self.left_key_list.get(0, tk.END)
        state["left_key_list"]["selection"] = self.left_key_list.curselection()
        state["right_key_list"] = {}
        state["right_key_list"]["contents"] = self.right_key_list.get(0, tk.END)
        state["right_key_list"]["selection"] = self.right_key_list.curselection()
        return state

    def load_state(self, state):
        # Update the row to match the state provided.
        self.name.insert(0, state["name"])
        self.left_table.insert(0, state["left_table"])
        self.right_table.insert(0, state["right_table"])
        if len(state["join_types"]) != 0:
            self.join_types.selection_set(state["join_types"][0])

        self.left_key_list.insert(0, *state["left_key_list"]["contents"])
        for index in state["left_key_list"]["selection"]:
            self.left_key_list.selection_set(index)

        self.right_key_list.insert(0, *state["right_key_list"]["contents"])
        for index in state["right_key_list"]["selection"]:
            self.right_key_list.selection_set(index)
